import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.contrib.postgres.search import SearchVectorField
from django.db import migrations, models

import misago.threads.models.attachment


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("misago_categories", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Post",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("poster_name", models.CharField(max_length=255)),
                ("poster_ip", models.GenericIPAddressField()),
                ("original", models.TextField()),
                ("parsed", models.TextField()),
                ("checksum", models.CharField(max_length=64, default="-")),
                ("attachments_cache", models.JSONField(null=True, blank=True)),
                ("posted_on", models.DateTimeField()),
                ("updated_on", models.DateTimeField()),
                ("edits", models.PositiveIntegerField(default=0)),
                (
                    "last_editor_name",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                (
                    "last_editor_slug",
                    models.SlugField(max_length=255, null=True, blank=True),
                ),
                (
                    "hidden_by",
                    models.ForeignKey(
                        related_name="+",
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                    ),
                ),
                (
                    "hidden_by_name",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                (
                    "hidden_by_slug",
                    models.SlugField(max_length=255, null=True, blank=True),
                ),
                ("hidden_on", models.DateTimeField(default=django.utils.timezone.now)),
                ("has_reports", models.BooleanField(default=False)),
                ("has_open_reports", models.BooleanField(default=False)),
                ("is_unapproved", models.BooleanField(default=False, db_index=True)),
                ("is_hidden", models.BooleanField(default=False)),
                ("is_protected", models.BooleanField(default=False)),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.Category",
                    ),
                ),
                (
                    "last_editor",
                    models.ForeignKey(
                        related_name="+",
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                    ),
                ),
                (
                    "mentions",
                    models.ManyToManyField(
                        related_name="mention_set", to=settings.AUTH_USER_MODEL
                    ),
                ),
                (
                    "poster",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                    ),
                ),
                ("is_event", models.BooleanField(default=False, db_index=True)),
                ("event_type", models.CharField(max_length=255, null=True, blank=True)),
                ("event_context", models.JSONField(null=True, blank=True)),
                ("likes", models.PositiveIntegerField(default=0)),
                ("last_likes", models.JSONField(blank=True, null=True)),
                ("search_document", models.TextField(blank=True, null=True)),
                ("search_vector", SearchVectorField()),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Thread",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                ("slug", models.CharField(max_length=255)),
                ("replies", models.PositiveIntegerField(default=0, db_index=True)),
                ("has_events", models.BooleanField(default=False)),
                ("has_poll", models.BooleanField(default=False)),
                ("has_reported_posts", models.BooleanField(default=False)),
                ("has_open_reports", models.BooleanField(default=False)),
                ("has_unapproved_posts", models.BooleanField(default=False)),
                ("has_hidden_posts", models.BooleanField(default=False)),
                ("started_on", models.DateTimeField(db_index=True)),
                ("starter_name", models.CharField(max_length=255)),
                ("starter_slug", models.CharField(max_length=255)),
                ("last_post_is_event", models.BooleanField(default=False)),
                ("last_post_on", models.DateTimeField(db_index=True)),
                (
                    "last_poster_name",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                (
                    "last_poster_slug",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                ("weight", models.PositiveIntegerField(default=0)),
                ("is_unapproved", models.BooleanField(default=False, db_index=True)),
                ("is_hidden", models.BooleanField(default=False)),
                ("is_closed", models.BooleanField(default=False)),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ThreadParticipant",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "thread",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.Thread",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("is_owner", models.BooleanField(default=False)),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name="thread",
            name="participants",
            field=models.ManyToManyField(
                related_name="privatethread_set",
                through="misago_threads.ThreadParticipant",
                to=settings.AUTH_USER_MODEL,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="post",
            name="thread",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="misago_threads.Thread"
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="thread",
            name="first_post",
            field=models.ForeignKey(
                related_name="+",
                on_delete=django.db.models.deletion.SET_NULL,
                blank=True,
                to="misago_threads.Post",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="thread",
            name="category",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="misago_categories.Category",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="thread",
            name="last_post",
            field=models.ForeignKey(
                related_name="+",
                on_delete=django.db.models.deletion.SET_NULL,
                blank=True,
                to="misago_threads.Post",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="thread",
            name="last_poster",
            field=models.ForeignKey(
                related_name="last_poster_set",
                on_delete=django.db.models.deletion.SET_NULL,
                blank=True,
                to=settings.AUTH_USER_MODEL,
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="thread",
            name="starter",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.SET_NULL,
                blank=True,
                to=settings.AUTH_USER_MODEL,
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AlterIndexTogether(
            name="thread",
            index_together=set(
                [
                    ("category", "id"),
                    ("category", "last_post_on"),
                    ("category", "replies"),
                ]
            ),
        ),
        migrations.AlterIndexTogether(
            name="post",
            index_together=set(
                [("thread", "id"), ("is_event", "is_hidden"), ("poster", "posted_on")]
            ),
        ),
        migrations.CreateModel(
            name="Subscription",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "last_read_on",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("send_email", models.BooleanField(default=False)),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.Category",
                    ),
                ),
                (
                    "thread",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.Thread",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.AlterIndexTogether(
            name="subscription", index_together=set([("send_email", "last_read_on")])
        ),
        migrations.CreateModel(
            name="PostEdit",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("edited_on", models.DateTimeField(default=django.utils.timezone.now)),
                ("editor_name", models.CharField(max_length=255)),
                ("editor_slug", models.CharField(max_length=255)),
                ("editor_ip", models.GenericIPAddressField()),
                ("edited_from", models.TextField()),
                ("edited_to", models.TextField()),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.Category",
                    ),
                ),
                (
                    "editor",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "post",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="edits_record",
                        to="misago_threads.Post",
                    ),
                ),
                (
                    "thread",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.Thread",
                    ),
                ),
            ],
            options={"ordering": ["-id"]},
        ),
        migrations.CreateModel(
            name="Attachment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("secret", models.CharField(max_length=64)),
                (
                    "uploaded_on",
                    models.DateTimeField(
                        default=django.utils.timezone.now, db_index=True
                    ),
                ),
                ("uploader_name", models.CharField(max_length=255)),
                ("uploader_slug", models.CharField(max_length=255, db_index=True)),
                ("uploader_ip", models.GenericIPAddressField()),
                ("filename", models.CharField(max_length=255, db_index=True)),
                ("size", models.PositiveIntegerField(default=0, db_index=True)),
                (
                    "thumbnail",
                    models.ImageField(
                        max_length=255,
                        blank=True,
                        null=True,
                        upload_to=misago.threads.models.attachment.upload_to,
                    ),
                ),
                (
                    "image",
                    models.ImageField(
                        max_length=255,
                        blank=True,
                        null=True,
                        upload_to=misago.threads.models.attachment.upload_to,
                    ),
                ),
                (
                    "file",
                    models.FileField(
                        max_length=255,
                        blank=True,
                        null=True,
                        upload_to=misago.threads.models.attachment.upload_to,
                    ),
                ),
                (
                    "post",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="misago_threads.Post",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AttachmentType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("extensions", models.CharField(max_length=255)),
                ("mimetypes", models.CharField(blank=True, max_length=255, null=True)),
                ("size_limit", models.PositiveIntegerField(default=1024)),
                (
                    "status",
                    models.PositiveIntegerField(
                        choices=[
                            (0, "Allow uploads and downloads"),
                            (1, "Allow downloads only"),
                            (2, "Disallow both uploading and downloading"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "limit_downloads_to",
                    models.ManyToManyField(
                        blank=True,
                        related_name="_attachmenttype_limit_downloads_to_+",
                        to="misago_acl.Role",
                    ),
                ),
                (
                    "limit_uploads_to",
                    models.ManyToManyField(
                        blank=True,
                        related_name="_attachmenttype_limit_uploads_to_+",
                        to="misago_acl.Role",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="attachment",
            name="filetype",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="misago_threads.AttachmentType",
            ),
        ),
        migrations.AddField(
            model_name="attachment",
            name="uploader",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.CreateModel(
            name="Poll",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("poster_name", models.CharField(max_length=255)),
                ("poster_slug", models.CharField(max_length=255)),
                ("poster_ip", models.GenericIPAddressField()),
                ("posted_on", models.DateTimeField(default=django.utils.timezone.now)),
                ("length", models.PositiveIntegerField(default=0)),
                ("question", models.CharField(max_length=255)),
                ("choices", models.JSONField()),
                ("allowed_choices", models.PositiveIntegerField(default=1)),
                ("allow_revotes", models.BooleanField(default=False)),
                ("votes", models.PositiveIntegerField(default=0)),
                ("is_public", models.BooleanField(default=False)),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.Category",
                    ),
                ),
                (
                    "poster",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "thread",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.Thread",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="PollVote",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("voter_name", models.CharField(max_length=255)),
                ("voter_slug", models.CharField(max_length=255)),
                ("voter_ip", models.GenericIPAddressField()),
                ("voted_on", models.DateTimeField(default=django.utils.timezone.now)),
                ("choice_hash", models.CharField(db_index=True, max_length=12)),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.Category",
                    ),
                ),
                (
                    "poll",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.Poll",
                    ),
                ),
                (
                    "thread",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.Thread",
                    ),
                ),
                (
                    "voter",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AlterIndexTogether(
            name="pollvote", index_together=set([("poll", "voter_name")])
        ),
        migrations.CreateModel(
            name="PostLike",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("liker_name", models.CharField(max_length=255, db_index=True)),
                ("liker_slug", models.CharField(max_length=255)),
                ("liker_ip", models.GenericIPAddressField()),
                ("liked_on", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.Category",
                    ),
                ),
            ],
            options={"ordering": ["-id"]},
        ),
        migrations.AddField(
            model_name="postlike",
            name="post",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="misago_threads.Post"
            ),
        ),
        migrations.AddField(
            model_name="postlike",
            name="thread",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="misago_threads.Thread"
            ),
        ),
        migrations.AddField(
            model_name="postlike",
            name="liker",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="post",
            name="liked_by",
            field=models.ManyToManyField(
                related_name="liked_post_set",
                through="misago_threads.PostLike",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from django.urls import reverse

from ..models import ThreadParticipant


def test_notify_about_new_reply_task_is_triggered_by_new_thread_reply(
    notify_on_new_thread_reply_mock, user_client, thread
):
    response = user_client.post(
        reverse("misago:api:thread-post-list", kwargs={"thread_pk": thread.pk}),
        data={
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200

    data = response.json()
    notify_on_new_thread_reply_mock.delay.assert_called_once_with(data["id"])


def test_notify_about_new_reply_task_is_triggered_by_new_private_thread_reply(
    notify_on_new_thread_reply_mock,
    user,
    user_client,
    private_thread,
):
    ThreadParticipant.objects.create(thread=private_thread, user=user)

    response = user_client.post(
        reverse(
            "misago:api:private-thread-post-list",
            kwargs={"thread_pk": private_thread.pk},
        ),
        data={
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200

    data = response.json()
    notify_on_new_thread_reply_mock.delay.assert_called_once_with(data["id"])


def test_notify_about_new_reply_task_is_not_triggered_on_new_thread_start(
    notify_on_new_thread_reply_mock, user_client, default_category
):
    response = user_client.post(
        reverse("misago:api:thread-list"),
        data={
            "category": default_category.pk,
            "title": "Test thread",
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200

    notify_on_new_thread_reply_mock.delay.assert_not_called()


def test_notify_about_new_reply_task_is_not_triggered_on_new_private_thread_start(
    notify_on_new_private_thread_mock,
    notify_on_new_thread_reply_mock,
    other_user,
    user_client,
):
    response = user_client.post(
        reverse("misago:api:private-thread-list"),
        data={
            "to": [other_user.username],
            "title": "Test thread",
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200

    notify_on_new_thread_reply_mock.delay.assert_not_called()


def test_notify_about_new_reply_task_is_not_triggered_on_thread_reply_edit(
    notify_on_new_thread_reply_mock,
    user_client,
    thread,
    user_reply,
):
    response = user_client.put(
        reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": thread.id, "pk": user_reply.id},
        ),
        json={
            "post": "Edited reply",
        },
    )

    assert response.status_code == 200

    notify_on_new_thread_reply_mock.delay.assert_not_called()


def test_notify_about_new_reply_task_is_not_triggered_on_private_thread_reply_edit(
    notify_on_new_thread_reply_mock,
    user,
    user_client,
    private_thread,
    private_thread_user_reply,
):
    ThreadParticipant.objects.create(thread=private_thread, user=user)

    response = user_client.put(
        reverse(
            "misago:api:private-thread-post-detail",
            kwargs={"thread_pk": private_thread.id, "pk": private_thread_user_reply.id},
        ),
        json={
            "post": "Edited reply",
        },
    )

    assert response.status_code == 200

    notify_on_new_thread_reply_mock.delay.assert_not_called()



# Generated by Django 2.2.1 on 2019-05-18 16:31

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("misago_users", "0018_auto_20190511_2051")]

    operations = [
        migrations.RunSQL("DROP INDEX IF EXISTS misago_user_is_staf_bf68aa_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_user_require_05204a_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_user_is_dele_2798b0_part"),
    ]



from unittest.mock import Mock

from django.urls import reverse

from ..views import get_protected_namespace

django_admin_url = reverse("admin:index")
misago_admin_url = reverse("misago:admin:index")
site_url = reverse("misago:index")


def test_request_to_misago_admin_root_url_is_protected():
    request = Mock(path=misago_admin_url)
    assert get_protected_namespace(request) == "misago:admin"


def test_request_to_misago_admin_subpath_url_is_protected():
    request = Mock(path=misago_admin_url + "users/")
    assert get_protected_namespace(request) == "misago:admin"


def test_request_to_django_admin_root_url_is_protected():
    request = Mock(path=django_admin_url)
    assert get_protected_namespace(request) == "admin"


def test_request_to_django_admin_subpath_url_is_protected():
    request = Mock(path=django_admin_url + "users/")
    assert get_protected_namespace(request) == "admin"


def test_request_to_site_root_url_is_not_protected():
    request = Mock(path=site_url)
    assert get_protected_namespace(request) is None


def test_request_to_site_subpath_url_is_not_protected():
    request = Mock(path=site_url + "t/some-thread-123/")
    assert get_protected_namespace(request) is None


def test_request_to_site_non_reversable_url_is_not_protected():
    request = Mock(path=site_url + "somewhere-custom/")
    assert get_protected_namespace(request) is None



from django.db import models
from django.utils.translation import pgettext_lazy


class MenuItem(models.Model):
    MENU_BOTH = "both"
    MENU_NAVBAR = "navbar"
    MENU_FOOTER = "footer"
    MENU_CHOICES = [
        (MENU_BOTH, pgettext_lazy("menu choice", "Navbar and footer")),
        (MENU_NAVBAR, pgettext_lazy("menu choice", "Navbar")),
        (MENU_FOOTER, pgettext_lazy("menu choice", "Footer")),
    ]

    menu = models.CharField(max_length=6, choices=MENU_CHOICES)
    title = models.CharField(max_length=255)
    url = models.URLField()
    order = models.IntegerField(default=0)
    css_class = models.CharField(max_length=255, null=True, blank=True)
    target_blank = models.BooleanField(default=False)
    rel = models.CharField(max_length=255, null=True, blank=True)

    class Meta:
        ordering = ("order",)
        get_latest_by = "order"

    def __str__(self):
        return self.title



from django.urls import reverse

from ..users.test import SuperUserTestCase


class AdminTestCase(SuperUserTestCase):
    def setUp(self):
        super().setUp()
        self.login_admin(self.user)

    def login_admin(self, user):
        self.client.post(
            reverse("misago:admin:index"),
            data={"username": user.email, "password": self.USER_PASSWORD},
        )
        self.client.get(reverse("misago:admin:index"))



from ..englishcorpus import EnglishCorpus


def test_corpus_has_length():
    corpus = EnglishCorpus()
    assert corpus


def test_corpus_can_be_shuffled():
    corpus = EnglishCorpus()
    corpus.shuffle()


def test_corpus_can_be_limited_to_phrases_shorter_than_specified():
    corpus = EnglishCorpus(max_length=100)
    assert corpus


def test_corpus_can_be_limited_to_phrases_longer_than_specified():
    corpus = EnglishCorpus(min_length=100)
    assert corpus


def test_corpus_produces_random_sequence():
    corpus = EnglishCorpus()
    choices = [corpus.random_sentence() for _ in range(2)]
    assert len(choices) == len(set(choices))


def test_corpus_produces_list_of_random_sentences():
    corpus = EnglishCorpus()
    assert len(corpus.random_sentences(5)) == 5



from logging import getLogger

from celery import shared_task
from django.conf import settings
from django.contrib.auth import get_user_model

from ..cache.versions import get_cache_versions
from ..conf.dynamicsettings import DynamicSettings
from ..core.pgutils import chunk_queryset
from ..users.bans import get_user_ban
from ..threads.models import Post, Thread
from .models import WatchedThread
from .threads import (
    notify_participant_on_new_private_thread,
    notify_watcher_on_new_thread_reply,
)

NOTIFY_CHUNK_SIZE = 20

User = get_user_model()
logger = getLogger("misago.notifications")


@shared_task(
    name="notifications.new-thread-reply",
    autoretry_for=(Post.DoesNotExist,),
    default_retry_delay=settings.MISAGO_NOTIFICATIONS_RETRY_DELAY,
    serializer="json",
)
def notify_on_new_thread_reply(reply_id: int):
    post = Post.objects.select_related("poster", "thread", "category").get(id=reply_id)
    post.thread.category = post.category

    if not WatchedThread.objects.filter(thread=post.thread).exists():
        return  # Nobody is watching this thread, stop

    cache_versions = get_cache_versions()
    dynamic_settings = DynamicSettings(cache_versions)

    queryset = WatchedThread.objects.filter(thread=post.thread).select_related("user")

    for watched_thread in chunk_queryset(queryset, NOTIFY_CHUNK_SIZE):
        if (
            watched_thread.user == post.poster
            or not watched_thread.user.is_active
            or get_user_ban(watched_thread.user, cache_versions)
        ):
            continue  # Skip poster and banned or inactive watchers

        try:
            notify_watcher_on_new_thread_reply(
                watched_thread, post, cache_versions, dynamic_settings
            )
        except Exception:
            logger.exception("Unexpected error in 'notify_watcher_on_new_thread_reply'")


@shared_task(
    name="notifications.new-private-thread",
    autoretry_for=(Thread.DoesNotExist,),
    default_retry_delay=settings.MISAGO_NOTIFICATIONS_RETRY_DELAY,
    serializer="json",
)
def notify_on_new_private_thread(
    actor_id: int,
    thread_id: int,
    participants: list[int],
):
    actor = User.objects.filter(id=actor_id).first()
    if not actor:
        return

    thread = Thread.objects.select_related("category", "first_post").get(id=thread_id)

    cache_versions = get_cache_versions()
    dynamic_settings = DynamicSettings(cache_versions)

    queryset = User.objects.filter(id__in=participants)

    for participant in chunk_queryset(queryset, NOTIFY_CHUNK_SIZE):
        if not participant.is_active or get_user_ban(participant, cache_versions):
            continue  # Skip inactive or banned participants

        try:
            notify_participant_on_new_private_thread(
                participant, actor, thread, cache_versions, dynamic_settings
            )
        except Exception:
            logger.exception(
                "Unexpected error in 'notify_participant_on_new_private_thread'"
            )


@shared_task(serializer="json")
def delete_duplicate_watched_threads(thread_id: int):
    # Merge send emails preference for watched threads
    email_notifications_users_ids = WatchedThread.objects.filter(
        thread_id=thread_id,
        send_emails=True,
    ).values("user_id")

    WatchedThread.objects.filter(
        thread_id=thread_id,
        send_emails=False,
        user_id__in=email_notifications_users_ids,
    ).update(send_emails=True)

    # Delete duplicate watched threads
    kept_watched_threads_ids = (
        WatchedThread.objects.filter(
            thread_id=thread_id,
        )
        .order_by("user_id", "-read_at")
        .distinct("user_id")
        .values("id")
    )

    WatchedThread.objects.filter(thread_id=thread_id).exclude(
        id__in=kept_watched_threads_ids
    ).delete()



# Generated by Django 4.2.7 on 2023-11-13 19:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("misago_users", "0024_user_notifications"),
    ]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="rank",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="misago_users.rank",
            ),
        ),
    ]



from unittest.mock import patch

import pytest
from django.urls import reverse

from ...notifications.models import Notification
from .. import test
from ...categories.models import Category
from ...readtracker import poststracker
from ..models import Poll, PollVote, Thread
from ..test import patch_category_acl, patch_other_category_acl
from .test_threads_api import ThreadsApiTestCase


class ThreadMergeApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        Category(name="Other Category", slug="other-category").insert_at(
            self.category, position="last-child", save=True
        )
        self.other_category = Category.objects.get(slug="other-category")

        self.api_link = reverse(
            "misago:api:thread-merge", kwargs={"pk": self.thread.pk}
        )

    @patch_category_acl({"can_merge_threads": False})
    def test_merge_no_permission(self):
        """api validates if thread can be merged with other one"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't merge threads in this category."}
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_no_url(self):
        """api validates if thread url was given"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Enter link to new thread."})

    @patch_category_acl({"can_merge_threads": True})
    def test_invalid_url(self):
        """api validates thread url"""
        response = self.client.post(
            self.api_link, {"other_thread": self.user.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "This is not a valid thread link."}
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_current_other_thread(self):
        """api validates if thread url given is to current thread"""
        response = self.client.post(
            self.api_link, {"other_thread": self.thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "You can't merge thread with itself."}
        )

    @patch_other_category_acl()
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_exists(self):
        """api validates if other thread exists"""
        other_thread = test.post_thread(self.other_category)
        other_other_thread = other_thread.get_absolute_url()
        other_thread.delete()

        response = self.client.post(self.api_link, {"other_thread": other_other_thread})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "The thread you have entered link to doesn't exist "
                    "or you don't have permission to see it."
                )
            },
        )

    @patch_other_category_acl({"can_see": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_is_invisible(self):
        """api validates if other thread is visible"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "The thread you have entered link to doesn't exist "
                    "or you don't have permission to see it."
                )
            },
        )

    @patch_other_category_acl({"can_merge_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_is_not_mergeable(self):
        """api validates if other thread can be merged"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "Other thread can't be merged with."}
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_close_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_thread_category_is_closed(self):
        """api validates if thread's category is open"""
        other_thread = test.post_thread(self.other_category)

        self.category.is_closed = True
        self.category.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't merge it's threads."},
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_close_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_thread_is_closed(self):
        """api validates if thread is open"""
        other_thread = test.post_thread(self.other_category)

        self.thread.is_closed = True
        self.thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't merge it with other threads."},
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_close_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_category_is_closed(self):
        """api validates if other thread's category is open"""
        other_thread = test.post_thread(self.other_category)

        self.other_category.is_closed = True
        self.other_category.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Other thread's category is closed. You can't merge with it."},
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_close_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_is_closed(self):
        """api validates if other thread is open"""
        other_thread = test.post_thread(self.other_category)

        other_thread.is_closed = True
        other_thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Other thread is closed and can't be merged with."},
        )

    @patch_other_category_acl({"can_merge_threads": True, "can_reply_threads": False})
    @patch_category_acl({"can_merge_threads": True})
    def test_other_thread_isnt_replyable(self):
        """api validates if other thread can be replied, which is condition for merge"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You can't merge this thread into thread you can't reply."},
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_merge_threads(self, delete_duplicate_watched_threads_mock):
        """api merges two threads successfully"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_merge_threads_kept_reads(self, delete_duplicate_watched_threads_mock):
        """api keeps both threads readtrackers after merge"""
        other_thread = test.post_thread(self.other_category)

        poststracker.save_read(self.user, self.thread.first_post)
        poststracker.save_read(self.user, other_thread.first_post)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # posts reads are kept
        postreads = self.user.postread_set.filter(post__is_event=False).order_by("id")

        self.assertEqual(
            list(postreads.values_list("post_id", flat=True)),
            [self.thread.first_post_id, other_thread.first_post_id],
        )
        self.assertEqual(postreads.filter(thread=other_thread).count(), 2)
        self.assertEqual(postreads.filter(category=self.other_category).count(), 2)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_merge_threads_kept_subs(self, delete_duplicate_watched_threads_mock):
        """api keeps other thread's subscription after merge"""
        other_thread = test.post_thread(self.other_category)

        self.user.subscription_set.create(
            thread=self.thread,
            category=self.thread.category,
            last_read_on=self.thread.last_post_on,
            send_email=False,
        )

        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=self.thread)
        self.user.subscription_set.get(category=self.category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # subscriptions are kept
        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_merge_threads_moved_subs(self, delete_duplicate_watched_threads_mock):
        """api keeps other thread's subscription after merge"""
        other_thread = test.post_thread(self.other_category)

        self.user.subscription_set.create(
            thread=other_thread,
            category=other_thread.category,
            last_read_on=other_thread.last_post_on,
            send_email=False,
        )

        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # subscriptions are kept
        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_merge_threads_handle_subs_conflict(
        self, delete_duplicate_watched_threads_mock
    ):
        """api resolves conflicting thread subscriptions after merge"""
        self.user.subscription_set.create(
            thread=self.thread,
            category=self.thread.category,
            last_read_on=self.thread.last_post_on,
            send_email=False,
        )

        other_thread = test.post_thread(self.other_category)

        self.user.subscription_set.create(
            thread=other_thread,
            category=other_thread.category,
            last_read_on=other_thread.last_post_on,
            send_email=False,
        )

        self.assertEqual(self.user.subscription_set.count(), 2)
        self.user.subscription_set.get(thread=self.thread)
        self.user.subscription_set.get(category=self.category)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # subscriptions are kept
        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=other_thread)
        self.user.subscription_set.get(category=self.other_category)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_merge_threads_kept_best_answer(
        self, delete_duplicate_watched_threads_mock
    ):
        """api merges two threads successfully, keeping best answer from old thread"""
        other_thread = test.post_thread(self.other_category)
        best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has three posts and an event now
        self.assertEqual(other_thread.post_set.count(), 4)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # best answer is kept in other thread
        other_thread = Thread.objects.get(pk=other_thread.pk)
        self.assertEqual(other_thread.best_answer, best_answer)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_merge_threads_moved_best_answer(
        self, delete_duplicate_watched_threads_mock
    ):
        """api merges two threads successfully, moving best answer to old thread"""
        other_thread = test.post_thread(self.other_category)

        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has three posts and an event now
        self.assertEqual(other_thread.post_set.count(), 4)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # best answer is kept in other thread
        other_thread = Thread.objects.get(pk=other_thread.pk)
        self.assertEqual(other_thread.best_answer, best_answer)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_merge_conflict_best_answer(self):
        """api errors on merge conflict, returning list of available best answers"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "best_answers": [
                    ["0", "Unmark all best answers"],
                    [str(self.thread.id), self.thread.title],
                    [str(other_thread.id), other_thread.title],
                ]
            },
        )

        # best answers were untouched
        self.assertEqual(self.thread.post_set.count(), 2)
        self.assertEqual(other_thread.post_set.count(), 2)
        self.assertEqual(
            Thread.objects.get(pk=self.thread.pk).best_answer_id, best_answer.id
        )
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, other_best_answer.id
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_best_answer_invalid_resolution(self):
        """api errors on invalid merge conflict resolution"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            {
                "other_thread": other_thread.get_absolute_url(),
                "best_answer": other_thread.id + 10,
            },
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Invalid choice."})

        # best answers were untouched
        self.assertEqual(self.thread.post_set.count(), 2)
        self.assertEqual(other_thread.post_set.count(), 2)
        self.assertEqual(
            Thread.objects.get(pk=self.thread.pk).best_answer_id, best_answer.id
        )
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, other_best_answer.id
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_threads_merge_conflict_unmark_all_best_answers(
        self, delete_duplicate_watched_threads_mock
    ):
        """api unmarks all best answers when unmark all choice is selected"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            {"other_thread": other_thread.get_absolute_url(), "best_answer": 0},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has four posts and an event now
        self.assertEqual(other_thread.post_set.count(), 5)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # final thread has no marked best answer
        self.assertIsNone(Thread.objects.get(pk=other_thread.pk).best_answer_id)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_threads_merge_conflict_keep_first_best_answer(
        self, delete_duplicate_watched_threads_mock
    ):
        """api unmarks other best answer on merge"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            {
                "other_thread": other_thread.get_absolute_url(),
                "best_answer": self.thread.pk,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has four posts and an event now
        self.assertEqual(other_thread.post_set.count(), 5)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # other thread's best answer was unchanged
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, best_answer.id
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_threads_merge_conflict_keep_other_best_answer(
        self, delete_duplicate_watched_threads_mock
    ):
        """api unmarks first best answer on merge"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.other_category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            {
                "other_thread": other_thread.get_absolute_url(),
                "best_answer": other_thread.pk,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has four posts and an event now
        self.assertEqual(other_thread.post_set.count(), 5)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # other thread's best answer was changed to merged in thread's answer
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, other_best_answer.id
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_merge_threads_kept_poll(self, delete_duplicate_watched_threads_mock):
        """api merges two threads successfully, keeping poll from other thread"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # poll and its votes were kept
        self.assertEqual(
            Poll.objects.filter(pk=poll.pk, thread=other_thread).count(), 1
        )
        self.assertEqual(
            PollVote.objects.filter(poll=poll, thread=other_thread).count(), 4
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_merge_threads_moved_poll(self, delete_duplicate_watched_threads_mock):
        """api merges two threads successfully, moving poll from old thread"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(self.thread, self.user)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # poll and its votes were moved
        self.assertEqual(
            Poll.objects.filter(pk=poll.pk, thread=other_thread).count(), 1
        )
        self.assertEqual(
            PollVote.objects.filter(poll=poll, thread=other_thread).count(), 4
        )

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_polls(self):
        """api errors on merge conflict, returning list of available polls"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(self.thread, self.user)
        other_poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "polls": [
                    ["0", "Delete all polls"],
                    [str(poll.pk), "%s (%s)" % (poll.question, poll.thread.title)],
                    [
                        str(other_poll.pk),
                        "%s (%s)" % (other_poll.question, other_poll.thread.title),
                    ],
                ]
            },
        )

        # polls and votes were untouched
        self.assertEqual(Poll.objects.count(), 2)
        self.assertEqual(PollVote.objects.count(), 8)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_poll_invalid_resolution(self):
        """api errors on invalid merge conflict resolution"""
        other_thread = test.post_thread(self.other_category)

        test.post_poll(self.thread, self.user)
        test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            {
                "other_thread": other_thread.get_absolute_url(),
                "poll": Poll.objects.all()[0].pk + 10,
            },
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Invalid choice."})

        # polls and votes were untouched
        self.assertEqual(Poll.objects.count(), 2)
        self.assertEqual(PollVote.objects.count(), 8)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_threads_merge_conflict_delete_all_polls(
        self, delete_duplicate_watched_threads_mock
    ):
        """api deletes all polls when delete all choice is selected"""
        other_thread = test.post_thread(self.other_category)
        test.post_poll(self.thread, self.user)
        test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link, {"other_thread": other_thread.get_absolute_url(), "poll": 0}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # polls and votes are gone
        self.assertEqual(Poll.objects.count(), 0)
        self.assertEqual(PollVote.objects.count(), 0)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_threads_merge_conflict_keep_first_poll(
        self, delete_duplicate_watched_threads_mock
    ):
        """api deletes other poll on merge"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(self.thread, self.user)
        other_poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            {"other_thread": other_thread.get_absolute_url(), "poll": poll.pk},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # other poll and its votes are gone
        self.assertEqual(Poll.objects.filter(thread=self.thread).count(), 0)
        self.assertEqual(PollVote.objects.filter(thread=self.thread).count(), 0)

        self.assertEqual(Poll.objects.filter(thread=other_thread).count(), 1)
        self.assertEqual(PollVote.objects.filter(thread=other_thread).count(), 4)

        Poll.objects.get(pk=poll.pk)
        with self.assertRaises(Poll.DoesNotExist):
            Poll.objects.get(pk=other_poll.pk)

    @patch_other_category_acl({"can_merge_threads": True})
    @patch_category_acl({"can_merge_threads": True})
    @patch("misago.threads.moderation.threads.delete_duplicate_watched_threads")
    def test_threads_merge_conflict_keep_other_poll(
        self, delete_duplicate_watched_threads_mock
    ):
        """api deletes first poll on merge"""
        other_thread = test.post_thread(self.other_category)
        poll = test.post_poll(self.thread, self.user)
        other_poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            {"other_thread": other_thread.get_absolute_url(), "poll": other_poll.pk},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": other_thread.id,
                "title": other_thread.title,
                "url": other_thread.get_absolute_url(),
            },
        )

        # other thread has two posts and an event now
        self.assertEqual(other_thread.post_set.count(), 3)

        # first thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # other poll and its votes are gone
        self.assertEqual(Poll.objects.filter(thread=self.thread).count(), 0)
        self.assertEqual(PollVote.objects.filter(thread=self.thread).count(), 0)

        self.assertEqual(Poll.objects.filter(thread=other_thread).count(), 1)
        self.assertEqual(PollVote.objects.filter(thread=other_thread).count(), 4)

        Poll.objects.get(pk=other_poll.pk)
        with self.assertRaises(Poll.DoesNotExist):
            Poll.objects.get(pk=poll.pk)


@pytest.fixture
def delete_duplicate_watched_threads_mock(mocker):
    return mocker.patch(
        "misago.threads.moderation.threads.delete_duplicate_watched_threads"
    )


@patch_category_acl({"can_merge_threads": True})
def test_thread_merge_api_merges_notifications(
    delete_duplicate_watched_threads_mock,
    user,
    user_client,
    thread,
    other_thread,
):
    notification = Notification.objects.create(
        user=user,
        verb="TEST",
        category=other_thread.category,
        thread=other_thread,
        thread_title=other_thread.title,
        post=other_thread.first_post,
    )

    response = user_client.post(
        reverse("misago:api:thread-merge", kwargs={"pk": thread.pk}),
        json={
            "other_thread": other_thread.get_absolute_url(),
        },
    )

    assert response.status_code == 200

    notification.refresh_from_db()

    assert notification.category_id == other_thread.category_id
    assert notification.thread_id == other_thread.id
    assert notification.thread_title == other_thread.title


@patch_category_acl({"can_merge_threads": True})
def test_thread_merge_api_merges_watched_threads(
    delete_duplicate_watched_threads_mock,
    user,
    user_client,
    thread,
    other_thread,
    watched_thread_factory,
):
    watched_thread = watched_thread_factory(user, thread, send_emails=True)

    response = user_client.post(
        reverse("misago:api:thread-merge", kwargs={"pk": thread.pk}),
        json={
            "other_thread": other_thread.get_absolute_url(),
        },
    )

    assert response.status_code == 200

    watched_thread.refresh_from_db()

    assert watched_thread.category_id == other_thread.category_id
    assert watched_thread.thread_id == other_thread.id

    delete_duplicate_watched_threads_mock.delay.assert_called_once_with(other_thread.id)



from django.core.cache import cache

from . import SETTINGS_CACHE
from ..cache.versions import invalidate_cache


def get_settings_cache(cache_versions):
    key = get_cache_key(cache_versions)
    return cache.get(key)


def set_settings_cache(cache_versions, user_settings):
    key = get_cache_key(cache_versions)
    cache.set(key, user_settings)


def get_cache_key(cache_versions):
    return "%s_%s" % (SETTINGS_CACHE, cache_versions[SETTINGS_CACHE])


def clear_settings_cache():
    invalidate_cache(SETTINGS_CACHE)



# Generated by Django 1.11.16 on 2018-11-25 15:15
from django.db import migrations, models

from .. import utils


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CacheVersion",
            fields=[
                (
                    "cache",
                    models.CharField(max_length=128, primary_key=True, serialize=False),
                ),
                (
                    "version",
                    models.CharField(
                        default=utils.generate_version_string, max_length=8
                    ),
                ),
            ],
        )
    ]



from datetime import timedelta

from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.template.defaultfilters import date as format_date
from django.utils import timezone
from django.utils.translation import pgettext_lazy

from ...acl import algebra
from ...acl.decorators import return_boolean
from ...acl.models import Role
from ...admin.forms import YesNoSwitch
from ..bans import get_user_ban

__all__ = [
    "allow_rename_user",
    "can_rename_user",
    "allow_moderate_avatar",
    "can_moderate_avatar",
    "allow_moderate_signature",
    "can_moderate_signature",
    "allow_edit_profile_details",
    "can_edit_profile_details",
    "allow_ban_user",
    "can_ban_user",
    "allow_lift_ban",
    "can_lift_ban",
]


class PermissionsForm(forms.Form):
    legend = pgettext_lazy("users moderation permission", "Users moderation")

    can_rename_users = YesNoSwitch(
        label=pgettext_lazy("users moderation permission", "Can rename users")
    )
    can_moderate_avatars = YesNoSwitch(
        label=pgettext_lazy("users moderation permission", "Can moderate avatars")
    )
    can_moderate_signatures = YesNoSwitch(
        label=pgettext_lazy("users moderation permission", "Can moderate signatures")
    )
    can_moderate_profile_details = YesNoSwitch(
        label=pgettext_lazy(
            "users moderation permission", "Can moderate profile details"
        )
    )
    can_ban_users = YesNoSwitch(
        label=pgettext_lazy("users moderation permission", "Can ban users")
    )
    max_ban_length = forms.IntegerField(
        label=pgettext_lazy(
            "users moderation permission", "Max length, in days, of imposed ban"
        ),
        help_text=pgettext_lazy(
            "users moderation permission",
            "Enter zero to let moderators impose permanent bans.",
        ),
        min_value=0,
        initial=0,
    )
    can_lift_bans = YesNoSwitch(
        label=pgettext_lazy("users moderation permission", "Can lift bans")
    )
    max_lifted_ban_length = forms.IntegerField(
        label=pgettext_lazy(
            "users moderation permission", "Max length, in days, of lifted ban"
        ),
        help_text=pgettext_lazy(
            "users moderation permission",
            "Enter zero to let moderators lift permanent bans.",
        ),
        min_value=0,
        initial=0,
    )


def change_permissions_form(role):
    if isinstance(role, Role) and role.special_role != "anonymous":
        return PermissionsForm


def build_acl(acl, roles, key_name):
    new_acl = {
        "can_rename_users": 0,
        "can_moderate_avatars": 0,
        "can_moderate_signatures": 0,
        "can_moderate_profile_details": 0,
        "can_ban_users": 0,
        "max_ban_length": 2,
        "can_lift_bans": 0,
        "max_lifted_ban_length": 2,
    }
    new_acl.update(acl)

    return algebra.sum_acls(
        new_acl,
        roles=roles,
        key=key_name,
        can_rename_users=algebra.greater,
        can_moderate_avatars=algebra.greater,
        can_moderate_signatures=algebra.greater,
        can_moderate_profile_details=algebra.greater,
        can_ban_users=algebra.greater,
        max_ban_length=algebra.greater_or_zero,
        can_lift_bans=algebra.greater,
        max_lifted_ban_length=algebra.greater_or_zero,
    )


def add_acl_to_user(user_acl, target):
    target.acl["can_rename"] = can_rename_user(user_acl, target)
    target.acl["can_moderate_avatar"] = can_moderate_avatar(user_acl, target)
    target.acl["can_moderate_signature"] = can_moderate_signature(user_acl, target)
    target.acl["can_edit_profile_details"] = can_edit_profile_details(user_acl, target)
    target.acl["can_ban"] = can_ban_user(user_acl, target)
    target.acl["max_ban_length"] = user_acl["max_ban_length"]
    target.acl["can_lift_ban"] = can_lift_ban(user_acl, target)

    mod_permissions = ["can_rename", "can_moderate_avatar", "can_moderate_signature"]

    for permission in mod_permissions:
        if target.acl[permission]:
            target.acl["can_moderate"] = True
            break


def register_with(registry):
    registry.acl_annotator(get_user_model(), add_acl_to_user)


def allow_rename_user(user_acl, target):
    if not user_acl["can_rename_users"]:
        raise PermissionDenied(
            pgettext_lazy("users moderation permission", "You can't rename users.")
        )
    if not user_acl["is_superuser"] and (target.is_staff or target.is_superuser):
        raise PermissionDenied(
            pgettext_lazy(
                "users moderation permission", "You can't rename administrators."
            )
        )


can_rename_user = return_boolean(allow_rename_user)


def allow_moderate_avatar(user_acl, target):
    if not user_acl["can_moderate_avatars"]:
        raise PermissionDenied(
            pgettext_lazy("users moderation permission", "You can't moderate avatars.")
        )
    if not user_acl["is_superuser"] and (target.is_staff or target.is_superuser):
        raise PermissionDenied(
            pgettext_lazy(
                "users moderation permission",
                "You can't moderate administrators avatars.",
            )
        )


can_moderate_avatar = return_boolean(allow_moderate_avatar)


def allow_moderate_signature(user_acl, target):
    if not user_acl["can_moderate_signatures"]:
        raise PermissionDenied(
            pgettext_lazy(
                "users moderation permission", "You can't moderate signatures."
            )
        )
    if not user_acl["is_superuser"] and (target.is_staff or target.is_superuser):
        message = pgettext_lazy(
            "users moderation permission",
            "You can't moderate administrators signatures.",
        )
        raise PermissionDenied(message)


can_moderate_signature = return_boolean(allow_moderate_signature)


def allow_edit_profile_details(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy(
                "users moderation permission",
                "You have to sign in to edit profile details.",
            )
        )
    if (
        user_acl["user_id"] != target.id
        and not user_acl["can_moderate_profile_details"]
    ):
        raise PermissionDenied(
            pgettext_lazy(
                "users moderation permission", "You can't edit other users details."
            )
        )
    if not user_acl["is_superuser"] and (target.is_staff or target.is_superuser):
        message = pgettext_lazy(
            "users moderation permission", "You can't edit administrators details."
        )
        raise PermissionDenied(message)


can_edit_profile_details = return_boolean(allow_edit_profile_details)


def allow_ban_user(user_acl, target):
    if not user_acl["can_ban_users"]:
        raise PermissionDenied(
            pgettext_lazy("users moderation permission", "You can't ban users.")
        )
    if target.is_staff or target.is_superuser:
        raise PermissionDenied(
            pgettext_lazy(
                "users moderation permission", "Administrators can't be banned."
            )
        )


can_ban_user = return_boolean(allow_ban_user)


def allow_lift_ban(user_acl, target):
    if not user_acl["can_lift_bans"]:
        raise PermissionDenied(
            pgettext_lazy("users moderation permission", "You can't lift bans.")
        )
    ban = get_user_ban(target, user_acl["cache_versions"])
    if not ban:
        raise PermissionDenied(
            pgettext_lazy("users moderation permission", "This user is not banned.")
        )
    if user_acl["max_lifted_ban_length"]:
        expiration_limit = timedelta(days=user_acl["max_lifted_ban_length"])
        lift_cutoff = (timezone.now() + expiration_limit).date()
        if not ban.valid_until:
            raise PermissionDenied(
                pgettext_lazy(
                    "users moderation permission", "You can't lift permanent bans."
                )
            )
        elif ban.valid_until > lift_cutoff:
            message = pgettext_lazy(
                "users moderation permission",
                "You can't lift bans that expire after %(expiration)s.",
            )
            raise PermissionDenied(message % {"expiration": format_date(lift_cutoff)})


can_lift_ban = return_boolean(allow_lift_ban)



# Generated by Django 1.11.16 on 2018-12-02 15:54
from django.db import migrations

from .. import MENU_ITEMS_CACHE
from ...cache.operations import StartCacheVersioning


class Migration(migrations.Migration):
    dependencies = [("misago_menus", "0001_initial"), ("misago_cache", "0001_initial")]

    operations = [StartCacheVersioning(MENU_ITEMS_CACHE)]



from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ... import MENU_ITEMS_CACHE


def test_top_menu_item_can_be_moved_down(admin_client, menu_item, other_menu_item):
    menu_item.order = 0
    menu_item.save()

    other_menu_item.order = 1
    other_menu_item.save()

    admin_client.post(
        reverse("misago:admin:settings:menu-items:down", kwargs={"pk": menu_item.pk})
    )

    menu_item.refresh_from_db()
    assert menu_item.order == 1
    other_menu_item.refresh_from_db()
    assert other_menu_item.order == 0


def test_top_menu_item_cant_be_moved_up(admin_client, menu_item, other_menu_item):
    menu_item.order = 0
    menu_item.save()

    other_menu_item.order = 1
    other_menu_item.save()

    admin_client.post(
        reverse("misago:admin:settings:menu-items:up", kwargs={"pk": menu_item.pk})
    )

    menu_item.refresh_from_db()
    assert menu_item.order == 0
    other_menu_item.refresh_from_db()
    assert other_menu_item.order == 1


def test_bottom_menu_item_cant_be_moved_down(admin_client, menu_item, other_menu_item):
    menu_item.order = 1
    menu_item.save()

    other_menu_item.order = 0
    other_menu_item.save()

    admin_client.post(
        reverse("misago:admin:settings:menu-items:down", kwargs={"pk": menu_item.pk})
    )

    menu_item.refresh_from_db()
    assert menu_item.order == 1
    other_menu_item.refresh_from_db()
    assert other_menu_item.order == 0


def test_bottom_menu_item_can_be_moved_up(admin_client, menu_item, other_menu_item):
    menu_item.order = 1
    menu_item.save()

    other_menu_item.order = 0
    other_menu_item.save()

    admin_client.post(
        reverse("misago:admin:settings:menu-items:up", kwargs={"pk": menu_item.pk})
    )

    menu_item.refresh_from_db()
    assert menu_item.order == 0
    other_menu_item.refresh_from_db()
    assert other_menu_item.order == 1


def test_moving_menu_item_down_invalidates_menu_items_cache(
    admin_client, menu_item, other_menu_item
):
    menu_item.order = 0
    menu_item.save()

    other_menu_item.order = 1
    other_menu_item.save()

    with assert_invalidates_cache(MENU_ITEMS_CACHE):
        admin_client.post(
            reverse(
                "misago:admin:settings:menu-items:down", kwargs={"pk": menu_item.pk}
            )
        )


def test_moving_menu_item_up_invalidates_menu_items_cache(
    admin_client, menu_item, other_menu_item
):
    menu_item.order = 1
    menu_item.save()

    other_menu_item.order = 0
    other_menu_item.save()

    with assert_invalidates_cache(MENU_ITEMS_CACHE):
        admin_client.post(
            reverse("misago:admin:settings:menu-items:up", kwargs={"pk": menu_item.pk})
        )



from django import forms
from django.utils.translation import pgettext_lazy

from .base import OAuthProviderForm


class TwitterForm(OAuthProviderForm):
    key = forms.CharField(
        label=pgettext_lazy("admin social auth twitter form", "Consumer API key"),
        required=False,
    )
    secret = forms.CharField(
        label=pgettext_lazy(
            "admin social auth twitter form", "Consumer API secret key"
        ),
        required=False,
    )



from .. import algebra


def test_greatest_value_is_returned():
    assert algebra.greater(1, 3) == 3
    assert algebra.greater(4, 2) == 4
    assert algebra.greater(2, 2) == 2
    assert algebra.greater(True, False) is True


def test_greatest_or_zero_value_is_returned():
    assert algebra.greater_or_zero(1, 3) == 3
    assert algebra.greater_or_zero(4, 2) == 4
    assert algebra.greater_or_zero(2, 2) == 2
    assert algebra.greater_or_zero(True, False) is False
    assert algebra.greater_or_zero(2, 0) == 0
    assert algebra.greater_or_zero(0, 0) == 0
    assert algebra.greater_or_zero(0, 120) == 0


def test_lower_value_is_returned():
    assert algebra.lower(1, 3) == 1
    assert algebra.lower(4, 2) == 2
    assert algebra.lower(2, 2) == 2
    assert algebra.lower(True, False) is False


def test_lowest_non_zero_value_is_returned():
    assert algebra.lower_non_zero(1, 3) == 1
    assert algebra.lower_non_zero(0, 2) == 2
    assert algebra.lower_non_zero(1, 2) == 1
    assert algebra.lower_non_zero(0, 0) == 0


def test_acls_are_be_added_together():
    test_acls = [
        {
            "can_see": 0,
            "can_hear": 0,
            "max_speed": 10,
            "min_age": 16,
            "speed_limit": 50,
        },
        {"can_see": 1, "can_hear": 0, "max_speed": 40, "min_age": 20, "speed_limit": 0},
        {
            "can_see": 0,
            "can_hear": 1,
            "max_speed": 80,
            "min_age": 18,
            "speed_limit": 40,
        },
    ]

    defaults = {
        "can_see": 0,
        "can_hear": 0,
        "max_speed": 30,
        "min_age": 18,
        "speed_limit": 60,
    }

    acl = algebra.sum_acls(
        defaults,
        acls=test_acls,
        can_see=algebra.greater,
        can_hear=algebra.greater,
        max_speed=algebra.greater,
        min_age=algebra.lower,
        speed_limit=algebra.greater_or_zero,
    )

    assert acl["can_see"] == 1
    assert acl["can_hear"] == 1
    assert acl["max_speed"] == 80
    assert acl["min_age"] == 16
    assert acl["speed_limit"] == 0



from django.contrib import messages
from django.utils.translation import pgettext_lazy

from ...admin.views import generic
from ..cache import clear_socialauth_cache
from ..models import SocialAuthProvider
from ..providers import providers


class SocialAuthProviderAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:settings:socialauth:index"
    model = SocialAuthProvider
    form_class = None
    templates_dir = "misago/admin/socialauth"
    message_404 = pgettext_lazy(
        "admin social auth providers", "Requested social login provider does not exist."
    )

    def get_target(self, kwargs):
        queryset = SocialAuthProvider.objects.filter(is_active=True)
        if self.is_atomic:
            queryset = queryset.select_for_update()
        return queryset.get(pk=kwargs["pk"])

    def get_queryset(self):
        return super().get_queryset().filter(is_active=True)


class SocialAuthProvidersList(SocialAuthProviderAdmin, generic.ListView):
    def process_context(self, request, context):
        active_providers = [i.pk for i in context["items"]]
        context["inactive_providers"] = []
        for provider in providers.list():
            if provider["provider"] not in active_providers:
                context["inactive_providers"].append(provider)
        return context


class EditSocialAuthProvider(SocialAuthProviderAdmin, generic.ModelFormView):
    def get_target(self, kwargs):
        try:
            return SocialAuthProvider.objects.get(provider=kwargs["pk"])
        except SocialAuthProvider.DoesNotExist:
            if not providers.is_registered(kwargs["pk"]):
                raise
            return SocialAuthProvider(provider=kwargs["pk"])

    def get_form_class(self, request, target):
        return providers.get_admin_form_class(target.provider)

    def get_form(self, form_class, request, target):
        if request.method == "POST":
            return form_class(request.POST, instance=target, request=request)
        return form_class(instance=target, initial=target.settings, request=request)

    def get_template_name(self, request, target):
        return providers.get_admin_template_name(target.provider)

    def handle_form(self, form, request, target):
        form.save()
        clear_socialauth_cache()

        message = pgettext_lazy(
            "admin social auth providers", "Login with %(provider)s has been updated."
        )
        messages.success(request, message % {"provider": target})


class DisableSocialAuthProvider(SocialAuthProviderAdmin, generic.ButtonView):
    def button_action(self, request, target):
        target.is_active = False
        target.save(update_fields=["is_active"])
        clear_socialauth_cache()

        message = pgettext_lazy(
            "admin social auth providers", "Login with %(provider)s has been disabled."
        )
        messages.success(request, message % {"provider": target})


class MoveDownSocialAuthProvider(SocialAuthProviderAdmin, generic.ButtonView):
    def button_action(self, request, target):
        try:
            other_target = SocialAuthProvider.objects.filter(
                is_active=True, order__gt=target.order
            )
            other_target = other_target.earliest("order")
        except SocialAuthProvider.DoesNotExist:
            other_target = None

        if other_target:
            other_target.order, target.order = target.order, other_target.order
            other_target.save(update_fields=["order"])
            target.save(update_fields=["order"])
            clear_socialauth_cache()

            message = pgettext_lazy(
                "admin social auth providers",
                "Login with %(provider)s has been moved after %(other)s.",
            )
            targets_names = {"provider": target, "other": other_target}
            messages.success(request, message % targets_names)


class MoveUpSocialAuthProvider(SocialAuthProviderAdmin, generic.ButtonView):
    def button_action(self, request, target):
        try:
            other_target = SocialAuthProvider.objects.filter(
                is_active=True, order__lt=target.order
            )
            other_target = other_target.latest("order")
        except SocialAuthProvider.DoesNotExist:
            other_target = None

        if other_target:
            other_target.order, target.order = target.order, other_target.order
            other_target.save(update_fields=["order"])
            target.save(update_fields=["order"])
            clear_socialauth_cache()

            message = pgettext_lazy(
                "admin social auth providers",
                "Login with %(provider)s has been moved before %(other)s.",
            )
            targets_names = {"provider": target, "other": other_target}
            messages.success(request, message % targets_names)



from unittest.mock import Mock

import pytest

from .. import exceptions
from ..client import SESSION_STATE, get_code_grant


def test_code_grant_is_returned_from_request():
    state = "l0r3m1p5um"
    code_grant = "valid-code"

    request = Mock(
        GET={
            "state": state,
            "code": code_grant,
        },
        session={SESSION_STATE: state},
    )

    assert get_code_grant(request) == code_grant

    # State was removed from session
    assert SESSION_STATE not in request.session


def test_exception_is_raised_if_provider_returned_error():
    request = Mock(
        GET={"error": "access_denied"},
        session={},
    )

    with pytest.raises(exceptions.OAuth2AccessDeniedError):
        get_code_grant(request)

    # State was removed from session
    assert SESSION_STATE not in request.session


def test_exception_is_raised_if_session_is_missing_state():
    state = "l0r3m1p5um"
    code_grant = "valid-code"

    request = Mock(
        GET={
            "state": state,
            "code": code_grant,
        },
        session={},
    )

    with pytest.raises(exceptions.OAuth2StateNotSetError):
        get_code_grant(request)


def test_exception_is_raised_if_request_is_missing_state():
    state = "l0r3m1p5um"
    code_grant = "valid-code"

    request = Mock(
        GET={
            "code": code_grant,
        },
        session={SESSION_STATE: state},
    )

    with pytest.raises(exceptions.OAuth2StateNotProvidedError):
        get_code_grant(request)

    # State was removed from session
    assert SESSION_STATE not in request.session


def test_exception_is_raised_if_request_state_is_empty():
    state = "l0r3m1p5um"
    code_grant = "valid-code"

    request = Mock(
        GET={
            "state": "",
            "code": code_grant,
        },
        session={SESSION_STATE: state},
    )

    with pytest.raises(exceptions.OAuth2StateNotProvidedError):
        get_code_grant(request)

    # State was removed from session
    assert SESSION_STATE not in request.session


def test_exception_is_raised_if_session_state_doesnt_match_with_request():
    state = "l0r3m1p5um"
    code_grant = "valid-code"

    request = Mock(
        GET={
            "state": "invalid",
            "code": code_grant,
        },
        session={SESSION_STATE: state},
    )

    with pytest.raises(exceptions.OAuth2StateMismatchError):
        get_code_grant(request)

    # State was removed from session
    assert SESSION_STATE not in request.session


def test_exception_is_raised_if_request_is_missing_code_grant():
    state = "l0r3m1p5um"

    request = Mock(
        GET={
            "state": state,
        },
        session={SESSION_STATE: state},
    )

    with pytest.raises(exceptions.OAuth2CodeNotProvidedError):
        get_code_grant(request)

    # State was removed from session
    assert SESSION_STATE not in request.session


def test_exception_is_raised_if_request_code_grant_is_empty():
    state = "l0r3m1p5um"

    request = Mock(
        GET={
            "code": "",
            "state": state,
        },
        session={SESSION_STATE: state},
    )

    with pytest.raises(exceptions.OAuth2CodeNotProvidedError):
        get_code_grant(request)

    # State was removed from session
    assert SESSION_STATE not in request.session



from django.dispatch import Signal, receiver

from ..users.signals import anonymize_user_data, username_changed
from .models import Category

delete_category_content = Signal()
move_category_content = Signal()


@receiver([anonymize_user_data, username_changed])
def update_usernames(sender, **kwargs):
    Category.objects.filter(last_poster=sender).update(
        last_poster_name=sender.username, last_poster_slug=sender.slug
    )



from collections import OrderedDict

from . import PostingMiddleware
from ....categories.models import Category


class SaveChangesMiddleware(PostingMiddleware):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.reset_state()

    def reset_state(self):
        self.user.update_all = False
        self.thread.update_all = False
        self.post.update_all = False

        self.user.update_fields = []
        self.thread.update_fields = []
        self.post.update_fields = []

        try:
            self.thread.category.update_all = False
            self.thread.category.update_fields = []
        except Category.DoesNotExist:
            # Exception for cases when thread has no category associated
            # If this is the case, its Category's middleware job to set those flags
            pass

    def save_models(self):
        self.save_model(self.user)
        self.save_model(self.thread.category)
        self.save_model(self.thread)
        self.save_model(self.post)
        self.reset_state()

    def save_model(self, model):
        if model.update_all:
            model.save()
        elif model.update_fields:
            update_fields = list(OrderedDict.fromkeys(model.update_fields))
            model.save(update_fields=update_fields)

    def save(self, serializer):
        self.save_models()

    def post_save(self, serializer):
        self.save_models()



import hashlib
import random

from django.contrib.auth import get_user_model
from django.utils.crypto import get_random_string

from ..users.bans import ban_user
from ..users.test import create_test_user
from .utils import retry_on_db_error

User = get_user_model()

AVATAR_SIZES = (400, 200, 100)
GRAVATAR_URL = "https://www.gravatar.com/avatar/%s?s=%s&d=retro"
PASSWORD = "password"


@retry_on_db_error
def get_fake_user(fake, rank=None, requires_activation=User.ACTIVATION_NONE):
    username = get_fake_username(fake)
    email = fake.email()
    return create_test_user(
        username,
        email.lower(),
        PASSWORD,
        avatars=get_fake_avatars(email),
        rank=rank,
        requires_activation=requires_activation,
    )


def get_fake_banned_user(fake, rank=None):
    user = get_fake_user(fake, rank=rank)
    ban_user(user)
    return user


def get_fake_inactive_user(fake, rank=None):
    return get_fake_user(fake, rank=rank, requires_activation=User.ACTIVATION_USER)


def get_fake_admin_activated_user(fake, rank=None):
    return get_fake_user(fake, rank=rank, requires_activation=User.ACTIVATION_ADMIN)


def get_fake_deleted_user(fake, rank=None):
    user = get_fake_user(fake, rank=rank)
    user.is_active = False
    user.save(update_fields=["is_active"])
    return user


def get_fake_username(fake):
    possible_usernames = [
        fake.first_name(),
        fake.last_name(),
        fake.name().replace(" ", ""),
        fake.user_name(),
        get_random_string(random.randint(4, 8)),
    ]

    return random.choice(possible_usernames)


def get_fake_avatars(email):
    email_hash = hashlib.md5(email.lower().encode()).hexdigest()

    return [
        {"size": size, "url": GRAVATAR_URL % (email_hash, size)}
        for size in AVATAR_SIZES
    ]



from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.shortcuts import redirect, render
from django.utils.translation import pgettext

from .searchproviders import searchproviders


def landing(request):
    allowed_providers = searchproviders.get_allowed_providers(request)
    if not request.user_acl["can_search"] or not allowed_providers:
        raise PermissionDenied(
            pgettext("search view", "You don't have permission to search site.")
        )

    default_provider = allowed_providers[0]
    return redirect("misago:search", search_provider=default_provider.url)


def search(request, search_provider):
    all_providers = searchproviders.get_providers(request)
    if not request.user_acl["can_search"] or not all_providers:
        raise PermissionDenied(
            pgettext("search view", "You don't have permission to search site.")
        )

    for provider in all_providers:
        if provider.url == search_provider:
            provider.allow_search()
            break
    else:
        raise Http404()

    if "q" in request.GET:
        search_query = request.GET.get("q").strip()
        request.frontend_context["SEARCH_QUERY"] = search_query
    else:
        search_query = ""

    return render(request, "misago/search.html", {"search_query": search_query})



from unittest.mock import Mock

import pytest

from ..exceptions import OAuth2UserDataValidationError
from ..validation import validate_user_data


def test_new_user_valid_data_is_validated(db, dynamic_settings):
    valid_data = validate_user_data(
        Mock(settings=dynamic_settings),
        None,
        {
            "id": "1234",
            "name": "Test User",
            "email": "user@example.com",
            "avatar": None,
        },
        {},
    )

    assert valid_data == {
        "id": "1234",
        "name": "Test_User",
        "email": "user@example.com",
        "avatar": None,
    }


def test_existing_user_valid_data_is_validated(user, dynamic_settings):
    valid_data = validate_user_data(
        Mock(settings=dynamic_settings),
        user,
        {
            "id": "1234",
            "name": user.username,
            "email": user.email,
            "avatar": None,
        },
        {},
    )

    assert valid_data == {
        "id": "1234",
        "name": user.username,
        "email": user.email,
        "avatar": None,
    }


def test_error_was_raised_for_user_data_with_without_name(
    db, dynamic_settings, disable_user_data_filters
):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        validate_user_data(
            Mock(settings=dynamic_settings),
            None,
            {
                "id": "1234",
                "name": "",
                "email": "user@example.com",
                "avatar": None,
            },
            {},
        )

    assert excinfo.value.error_list == [
        (
            "Username can only contain Latin alphabet letters, digits, "
            "and an underscore sign."
        )
    ]


def test_error_was_raised_for_user_data_with_invalid_name(
    db, dynamic_settings, disable_user_data_filters
):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        validate_user_data(
            Mock(settings=dynamic_settings),
            None,
            {
                "id": "1234",
                "name": "Invalid!",
                "email": "user@example.com",
                "avatar": None,
            },
            {},
        )

    assert excinfo.value.error_list == [
        (
            "Username can only contain Latin alphabet letters, digits, "
            "and an underscore sign."
        )
    ]


def test_error_was_raised_for_user_data_with_too_long_name(db, dynamic_settings):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        validate_user_data(
            Mock(settings=dynamic_settings),
            None,
            {
                "id": "1234",
                "name": "UserName" * 100,
                "email": "user@example.com",
                "avatar": None,
            },
            {},
        )

    assert excinfo.value.error_list == [
        "Username cannot be longer than 200 characters."
    ]


def test_error_was_raised_for_user_data_without_email(db, dynamic_settings):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        validate_user_data(
            Mock(settings=dynamic_settings),
            None,
            {
                "id": "1234",
                "name": "Test User",
                "email": "",
                "avatar": None,
            },
            {},
        )

    assert excinfo.value.error_list == ["Enter a valid e-mail address."]


def test_error_was_raised_for_user_data_with_invalid_email(db, dynamic_settings):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        validate_user_data(
            Mock(settings=dynamic_settings),
            None,
            {
                "id": "1234",
                "name": "Test User",
                "email": "userexample.com",
                "avatar": None,
            },
            {},
        )

    assert excinfo.value.error_list == ["Enter a valid e-mail address."]



# Generated by Django 2.2.3 on 2019-09-28 21:34

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [("misago_users", "0021_user_sso_id")]

    operations = [
        migrations.CreateModel(
            name="DeletedUser",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "deleted_on",
                    models.DateTimeField(
                        db_index=True, default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_by",
                    models.PositiveIntegerField(
                        choices=[(1, "By self"), (2, "By staff"), (3, "By system")],
                        db_index=True,
                    ),
                ),
            ],
            options={"ordering": ["-id"]},
        )
    ]



from rest_framework import serializers

from django.core.exceptions import PermissionDenied
from django.http import QueryDict
from django.utils import timezone
from django.utils.module_loading import import_string

from ....conf import settings


class PostingInterrupt(Exception):
    def __init__(self, message):  # pylint: disable=super-init-not-called
        if not message:
            raise ValueError("You have to provide PostingInterrupt message.")
        self.message = message


class PostingEndpoint:
    START = 0
    REPLY = 1
    EDIT = 2

    def __init__(self, request, mode, **kwargs):
        # we are using lock on user model to protect us from flood
        request.user.lock()

        # build kwargs dict for passing to middlewares
        self.kwargs = kwargs
        self.kwargs.update(
            {
                "mode": mode,
                "request": request,
                "settings": request.settings,
                "user": request.user,
                "user_acl": request.user_acl,
            }
        )

        self.__dict__.update(kwargs)

        # some middlewares (eg. emailnotification) may call render()
        # which will crash if this isn't set to false
        request.include_frontend_context = False

        self.datetime = timezone.now()
        self.errors = {}
        self._is_validated = False

        self.middlewares = self._load_middlewares()
        self._serializers = self._initialize_serializers()

    @property
    def is_start_endpoint(self):
        return self.mode == self.START

    @property
    def is_reply_endpoint(self):
        return self.mode == self.REPLY

    @property
    def is_edit_endpoint(self):
        return self.mode == self.EDIT

    def _load_middlewares(self):
        kwargs = self.kwargs.copy()
        kwargs.update({"datetime": self.datetime, "parsing_result": {}})

        middlewares = []
        for middleware in settings.MISAGO_POSTING_MIDDLEWARES:
            middleware_class = import_string(middleware)

            try:
                middleware_obj = middleware_class(prefix=middleware, **kwargs)
                if middleware_obj.use_this_middleware():
                    middlewares.append((middleware, middleware_obj))
            except PostingInterrupt:
                raise ValueError(
                    "Posting process can only be interrupted during pre_save phase"
                )

        return middlewares

    def get_serializers(self):
        """return list of serializers belonging to serializerset"""
        return self._serializers

    def _initialize_serializers(self):
        try:
            serializers = {}
            for middleware, obj in self.middlewares:
                serializer = obj.get_serializer()
                if serializer:
                    serializers[middleware] = serializer
            return serializers
        except PostingInterrupt:
            raise ValueError(
                "Posting process can only be interrupted during pre_save phase"
            )

    def is_valid(self):
        """validate data against all serializers"""
        for serializer in self._serializers.values():
            if not serializer.is_valid():
                self.errors.update(serializer.errors)

        self._is_validated = True
        return not self.errors

    def save(self):
        """save new state to backend"""
        if not self._is_validated or self.errors:
            raise RuntimeError(
                "You need to validate posting data successfully before calling save"
            )

        try:
            for middleware, obj in self.middlewares:
                obj.pre_save(self._serializers.get(middleware))
        except PostingInterrupt as e:
            raise ValueError(
                "Posting process can only be interrupted "
                "from within interrupt_posting method"
            )

        try:
            for middleware, obj in self.middlewares:
                obj.interrupt_posting(self._serializers.get(middleware))
        except PostingInterrupt as e:
            raise PermissionDenied(e.message)

        try:
            for middleware, obj in self.middlewares:
                obj.save(self._serializers.get(middleware))
            for middleware, obj in self.middlewares:
                obj.post_save(self._serializers.get(middleware))
        except PostingInterrupt as e:
            raise ValueError(
                "Posting process can only be interrupted "
                "from within interrupt_posting method"
            )


class PostingMiddleware:
    """abstract middleware class"""

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.__dict__.update(kwargs)

    def use_this_middleware(self):
        return True

    def get_serializer(self):
        pass

    def pre_save(self, serializer):
        pass

    def interrupt_posting(self, serializer):
        pass

    def save(self, serializer):
        pass

    def post_save(self, serializer):
        pass



from typing import List

from django.urls import reverse

from .models import Rank
from .pages import usercp, users_list


def get_users_menus(request) -> dict:
    return {
        "userOptions": get_user_options_pages(request),
        "usersLists": get_users_lists(request),
    }


def get_user_options_pages(request) -> List[dict]:
    links = []

    if not request.user.is_authenticated:
        return links

    for section in usercp.get_sections(request):
        links.append(
            {
                "icon": section["icon"],
                "name": str(section["name"]),
                "url": reverse(section["link"]),
            }
        )

    return links


def get_users_lists(request) -> List[dict]:
    links = []

    for section in users_list.get_sections(request):
        links.append({"name": str(section["name"]), "url": reverse(section["link"])})

    ranks_queryset = (
        Rank.objects.filter(is_tab=True).order_by("order").values("id", "name", "slug")
    )

    for rank in ranks_queryset:
        links.append(
            {
                "name": rank["name"],
                "url": reverse("misago:users-rank", kwargs={"slug": rank["slug"]}),
            }
        )

    return links



# Generated by Django 2.2.1 on 2019-05-19 00:08

from django.db import migrations, models
from django.db.models import F


def set_default_dry_value(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")
    Setting.objects.filter(dry_value__isnull=True, default_value__isnull=False).update(
        dry_value=F("default_value")
    )


class Migration(migrations.Migration):
    dependencies = [("misago_conf", "0002_cache_version")]

    operations = [
        migrations.RemoveField(model_name="setting", name="description"),
        migrations.RemoveField(model_name="setting", name="field_extra"),
        migrations.RemoveField(model_name="setting", name="form_field"),
        migrations.RemoveField(model_name="setting", name="group"),
        migrations.RemoveField(model_name="setting", name="legend"),
        migrations.RemoveField(model_name="setting", name="name"),
        migrations.RemoveField(model_name="setting", name="order"),
        migrations.RunPython(set_default_dry_value),
        migrations.RemoveField(model_name="setting", name="default_value"),
        migrations.DeleteModel(name="SettingsGroup"),
        migrations.AddField(
            model_name="setting",
            name="image",
            field=models.ImageField(
                blank=True,
                null=True,
                upload_to="conf",
                height_field="image_height",
                width_field="image_width",
            ),
        ),
        migrations.AddField(
            model_name="setting",
            name="image_size",
            field=models.PositiveIntegerField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name="setting",
            name="image_width",
            field=models.PositiveIntegerField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name="setting",
            name="image_height",
            field=models.PositiveIntegerField(null=True, blank=True),
        ),
    ]



from unittest.mock import Mock

from ..context_processors import preload_categories_json


def test_categories_json_is_included_in_frontend_json(
    cache_versions,
    user,
    default_category,
    sibling_category,
    child_category,
    other_category,
):
    request = Mock(
        include_frontend_context=True,
        frontend_context={},
        cache_versions=cache_versions,
        user=user,
        user_acl={
            "visible_categories": [
                default_category.id,
                sibling_category.id,
                child_category.id,
                other_category.id,
            ],
        },
    )

    preload_categories_json(request)

    assert request.frontend_context["categoriesMap"] == [
        {
            "id": default_category.id,
            "name": default_category.name,
            "shortName": default_category.short_name,
            "color": default_category.color,
            "url": default_category.get_absolute_url(),
        },
        {
            "id": sibling_category.id,
            "name": sibling_category.name,
            "shortName": sibling_category.short_name,
            "color": sibling_category.color,
            "url": sibling_category.get_absolute_url(),
        },
        {
            "id": other_category.id,
            "name": other_category.name,
            "shortName": other_category.short_name,
            "color": other_category.color,
            "url": other_category.get_absolute_url(),
        },
    ]



import os
from datetime import timedelta
from io import StringIO

from django.core.files import File
from django.core.management import call_command

from ..datadownloads import request_user_data_download
from ..management.commands import expireuserdatadownloads
from ..models import DataDownload
from ..test import AuthenticatedUserTestCase

TESTFILES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testfiles")
TEST_FILE_PATH = os.path.join(TESTFILES_DIR, "avatar.png")


class ExpireUserDataDownloadsTests(AuthenticatedUserTestCase):
    def test_delete_expired_data_download(self):
        """management command deletes expired data download"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_READY
        with open(TEST_FILE_PATH, "rb") as download_file:
            data_download.file = File(download_file)
            data_download.save()

        out = StringIO()
        call_command(expireuserdatadownloads.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Data downloads expired: 1")

        updated_data_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_data_download.status, DataDownload.STATUS_EXPIRED)
        self.assertFalse(updated_data_download.file)

    def test_skip_not_expired_data_download(self):
        """management command skips data download that expires in future"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_READY
        data_download.expires_on += timedelta(hours=1)
        with open(TEST_FILE_PATH, "rb") as download_file:
            data_download.file = File(download_file)
            data_download.save()

        out = StringIO()
        call_command(expireuserdatadownloads.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Data downloads expired: 0")

        updated_data_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_data_download.status, DataDownload.STATUS_READY)
        self.assertTrue(updated_data_download.file)

    def test_skip_pending_data_download(self):
        """management command skips pending data downloads"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_PENDING
        data_download.save()

        out = StringIO()
        call_command(expireuserdatadownloads.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Data downloads expired: 0")

        updated_data_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_data_download.status, DataDownload.STATUS_PENDING)

    def test_skip_processing_data_download(self):
        """management command skips processing data downloads"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_PROCESSING
        data_download.save()

        out = StringIO()
        call_command(expireuserdatadownloads.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Data downloads expired: 0")

        updated_data_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_data_download.status, DataDownload.STATUS_PROCESSING)

    def test_skip_expired_data_download(self):
        """management command skips pending data downloads"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_EXPIRED
        data_download.save()

        out = StringIO()
        call_command(expireuserdatadownloads.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Data downloads expired: 0")

        updated_data_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_data_download.status, DataDownload.STATUS_EXPIRED)

    def test_no_expired_data_download(self):
        """management command doesn't error when no data is expired"""
        out = StringIO()
        call_command(expireuserdatadownloads.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Data downloads expired: 0")



from django.contrib.auth import REDIRECT_FIELD_NAME, login
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.urls import reverse
from django.utils.translation import pgettext
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt
from social_core.actions import do_auth, do_complete
from social_django.views import _do_login
from social_django.utils import load_strategy


def get_provider_from_request(request, backend):
    try:
        return request.socialauth[backend]
    except KeyError:
        raise Http404()


def social_auth_view(f):
    def social_auth_view_wrapper(request, backend, *args, **kwargs):
        if request.settings.enable_oauth2_client:
            raise PermissionDenied(
                pgettext(
                    "social auth",
                    "This feature has been disabled. Please use %(provider)s to sign in.",
                )
                % {"provider": request.settings.oauth2_provider}
            )

        provider = get_provider_from_request(request, backend)
        request.strategy = load_strategy(request)

        backend_class = provider["auth_backend"]
        request.backend = backend_class(
            request.strategy,
            reverse("misago:social-complete", kwargs={"backend": backend}),
        )

        return f(request, backend, *args, **kwargs)

    return social_auth_view_wrapper


@never_cache
@social_auth_view
def auth(request, backend):
    return do_auth(request.backend, redirect_name=REDIRECT_FIELD_NAME)


@never_cache
@csrf_exempt
@social_auth_view
def complete(request, backend, *args, **kwargs):
    return do_complete(
        request.backend,
        do_login,
        user=request.user,
        redirect_name=REDIRECT_FIELD_NAME,
        request=request,
        *args,
        **kwargs
    )


def do_login(backend, user, social_user):
    user.backend = "misago.users.authbackends.MisagoBackend"
    login(backend.strategy.request, user)



from django.db import migrations

from ...core.utils import slugify

pgettext_lazy = lambda c, s: s


def create_default_ranks(apps, schema_editor):
    Rank = apps.get_model("misago_users", "Rank")

    team = Rank.objects.create(
        name=pgettext_lazy("rank name", "Forum team"),
        slug=slugify(pgettext_lazy("rank name", "Forum team")),
        title=pgettext_lazy("rank name", "Team"),
        css_class="primary",
        is_tab=True,
        order=0,
    )

    member = Rank.objects.create(
        name=pgettext_lazy("rank name", "Members"),
        slug=slugify(pgettext_lazy("rank name", "Members")),
        is_default=True,
        order=1,
    )

    Role = apps.get_model("misago_acl", "Role")

    team.roles.add(Role.objects.get(name=pgettext_lazy("role name", "Moderator")))
    team.roles.add(Role.objects.get(name=pgettext_lazy("role name", "Private threads")))
    team.roles.add(
        Role.objects.get(name=pgettext_lazy("role name", "Private threads moderator"))
    )
    team.roles.add(Role.objects.get(name=pgettext_lazy("role name", "Deleting users")))

    member.roles.add(
        Role.objects.get(name=pgettext_lazy("role name", "Private threads"))
    )


class Migration(migrations.Migration):
    dependencies = [
        ("misago_users", "0003_bans_version_tracker"),
        ("misago_acl", "0003_default_roles"),
    ]

    operations = [migrations.RunPython(create_default_ranks)]



from django.urls import path
from django.utils.translation import pgettext_lazy

from .views import (
    AnalyticsSettingsView,
    CaptchaSettingsView,
    GeneralSettingsView,
    NotificationsSettingsView,
    OAuth2SettingsView,
    ThreadsSettingsView,
    UsersSettingsView,
    index,
)


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        urlpatterns.namespace("settings/", "settings")

        urlpatterns.patterns("settings", path("", index, name="index"))

        urlpatterns.single_pattern(
            "analytics/",
            "analytics",
            "settings",
            AnalyticsSettingsView.as_view(),
        )
        urlpatterns.single_pattern(
            "captcha/", "captcha", "settings", CaptchaSettingsView.as_view()
        )
        urlpatterns.single_pattern(
            "general/", "general", "settings", GeneralSettingsView.as_view()
        )
        urlpatterns.single_pattern(
            "notifications/",
            "notifications",
            "settings",
            NotificationsSettingsView.as_view(),
        )
        urlpatterns.single_pattern(
            "oauth2/", "oauth2", "settings", OAuth2SettingsView.as_view()
        )
        urlpatterns.single_pattern(
            "threads/", "threads", "settings", ThreadsSettingsView.as_view()
        )
        urlpatterns.single_pattern(
            "users/", "users", "settings", UsersSettingsView.as_view()
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=pgettext_lazy("admin node", "Settings"),
            icon="fa fa-cog",
            after="themes:index",
            namespace="settings",
        )

        site.add_node(
            name=pgettext_lazy("admin node", "General"),
            description=pgettext_lazy(
                "admin node", "Change forum details like name, description or footer."
            ),
            parent="settings",
            namespace="general",
        )
        site.add_node(
            name=pgettext_lazy("admin node", "Users"),
            description=pgettext_lazy(
                "admin node",
                "Customize user accounts default behavior and features availability.",
            ),
            parent="settings",
            namespace="users",
            after="general:index",
        )
        site.add_node(
            name=pgettext_lazy("admin node", "Captcha"),
            description=pgettext_lazy(
                "admin node",
                "Setup protection against automatic registrations on the site.",
            ),
            parent="settings",
            namespace="captcha",
            after="users:index",
        )
        site.add_node(
            name=pgettext_lazy("admin node", "Threads"),
            description=pgettext_lazy(
                "admin node", "Threads, posts, polls and attachments options."
            ),
            parent="settings",
            namespace="threads",
            after="captcha:index",
        )
        site.add_node(
            name=pgettext_lazy("admin node", "Notifications"),
            description=pgettext_lazy(
                "admin node",
                "Those settings control default notification preferences of new user accounts and storage time of existing notifications.",
            ),
            parent="settings",
            namespace="notifications",
            after="threads:index",
        )
        site.add_node(
            name=pgettext_lazy("admin node", "OAuth2"),
            description=pgettext_lazy(
                "admin node",
                "Enable OAuth2 client and connect your site to existing auth provider.",
            ),
            parent="settings",
            namespace="oauth2",
            after="notifications:index",
        )
        site.add_node(
            name=pgettext_lazy("admin node", "Analytics"),
            description=pgettext_lazy(
                "admin node",
                "Enable Google Analytics or setup Google Site Verification.",
            ),
            parent="settings",
            namespace="analytics",
            after="oauth2:index",
        )



from django import forms
from django.utils.translation import pgettext_lazy

from ...admin.forms import YesNoSwitch
from ..models import MenuItem
from ..cache import clear_menus_cache


class MenuItemForm(forms.ModelForm):
    title = forms.CharField(label=pgettext_lazy("admin menu item form", "Title"))
    url = forms.URLField(
        label=pgettext_lazy("admin menu item form", "URL"),
        help_text=pgettext_lazy(
            "admin menu item form", "URL where this item will point to."
        ),
    )
    menu = forms.ChoiceField(
        label=pgettext_lazy("admin menu item form", "Menu"),
        choices=MenuItem.MENU_CHOICES,
        help_text=pgettext_lazy(
            "admin menu item form", "Menu in which this item will be displayed."
        ),
    )
    css_class = forms.CharField(
        label=pgettext_lazy("admin menu item form", "CSS class"),
        help_text=pgettext_lazy(
            "admin menu item form",
            'Optional. Additional CSS class to include in link\'s "class" HTML attribute.',
        ),
        required=False,
    )
    target_blank = YesNoSwitch(
        label=pgettext_lazy("admin menu item form", "Open this link in new window"),
        help_text=pgettext_lazy(
            "admin menu item form",
            'Enabling this option will result in the target="_blank" attribute being added to this link\'s HTML element.',
        ),
        required=False,
    )
    rel = forms.CharField(
        label=pgettext_lazy("admin menu item form", "Rel attribute"),
        help_text=pgettext_lazy(
            "admin menu item form",
            'Optional "rel" attribute that this item will use (ex. "nofollow").',
        ),
        required=False,
    )

    class Meta:
        model = MenuItem
        fields = ["title", "url", "menu", "css_class", "target_blank", "rel"]

    def save(self):
        item = super().save()
        clear_menus_cache()
        return item



from django.apps import AppConfig

from .providers import providers


class MisagoACLsConfig(AppConfig):
    name = "misago.acl"
    label = "misago_acl"
    verbose_name = "Misago ACL framework"

    def ready(self):
        providers.load()



from django.urls import path
from django.utils.translation import pgettext_lazy

from .views.categories import (
    CategoriesList,
    DeleteCategory,
    EditCategory,
    MoveDownCategory,
    MoveUpCategory,
    NewCategory,
)
from .views.perms import (
    CategoryPermissions,
    CategoryRolesList,
    DeleteCategoryRole,
    EditCategoryRole,
    NewCategoryRole,
    RoleCategoriesACL,
)


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # Categories section
        urlpatterns.namespace("categories/", "categories")

        # Nodes
        urlpatterns.patterns(
            "categories",
            path("", CategoriesList.as_view(), name="index"),
            path("new/", NewCategory.as_view(), name="new"),
            path("edit/<int:pk>/", EditCategory.as_view(), name="edit"),
            path(
                "permissions/<int:pk>/",
                CategoryPermissions.as_view(),
                name="permissions",
            ),
            path("move/down/<int:pk>/", MoveDownCategory.as_view(), name="down"),
            path("move/up/<int:pk>/", MoveUpCategory.as_view(), name="up"),
            path("delete/<int:pk>/", DeleteCategory.as_view(), name="delete"),
        )

        # Category Roles
        urlpatterns.namespace("categories/", "categories", "permissions")
        urlpatterns.patterns(
            "permissions:categories",
            path("", CategoryRolesList.as_view(), name="index"),
            path("new/", NewCategoryRole.as_view(), name="new"),
            path("edit/<int:pk>/", EditCategoryRole.as_view(), name="edit"),
            path("delete/<int:pk>/", DeleteCategoryRole.as_view(), name="delete"),
        )

        # Change Role Category Permissions
        urlpatterns.patterns(
            "permissions",
            path(
                "categories/<int:pk>/",
                RoleCategoriesACL.as_view(),
                name="categories",
            ),
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=pgettext_lazy("admin node", "Categories"),
            icon="fas fa-sitemap",
            after="ranks:index",
            namespace="categories",
        )

        site.add_node(
            name=pgettext_lazy("admin node", "Category permissions"),
            parent="permissions",
            namespace="categories",
        )



from io import StringIO

from django.core.management import call_command

from ...acl import ACL_CACHE
from ...cache.test import assert_invalidates_cache
from ..management.commands import createfakecategories


def test_management_command_creates_fake_categories(root_category):
    call_command(createfakecategories.Command(), categories=5, stdout=StringIO())
    root_category.refresh_from_db()
    assert root_category.get_descendant_count() == 6  # 5 fakes + 1 default


def test_management_command_updates_categories_tree_after_creation(root_category):
    call_command(createfakecategories.Command(), categories=5, stdout=StringIO())
    root_category.refresh_from_db()
    assert root_category.rght == root_category.lft + 13  # 6 child items


def test_management_command_creates_categories_at_specified_level(default_category):
    call_command(
        createfakecategories.Command(),
        categories=5,
        level=2,
        stdout=StringIO(),
    )

    default_category.refresh_from_db()
    assert default_category.get_descendant_count() == 5


def test_management_command_copies_default_category_acl(default_category):
    call_command(
        createfakecategories.Command(),
        categories=5,
        level=1,
        stdout=StringIO(),
    )

    default_category.refresh_from_db()
    default_acls_count = default_category.category_role_set.count()

    for fake_category in default_category.get_descendants():
        assert fake_category.category_role_set.count() == default_acls_count


def test_management_command_invalidates_acl_cache(db):
    with assert_invalidates_cache(ACL_CACHE):
        call_command(createfakecategories.Command(), categories=5, stdout=StringIO())



from datetime import timedelta
from unittest.mock import Mock

from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ..cutoffdate import get_cutoff_date


def test_cutoff_date_for_no_user_is_calculated_from_setting(dynamic_settings):
    cutoff_date = get_cutoff_date(dynamic_settings)
    valid_cutoff_date = timezone.now() - timedelta(
        days=dynamic_settings.readtracker_cutoff
    )
    assert cutoff_date < valid_cutoff_date


def test_cutoff_date_for_recently_joined_user_is_their_join_date(dynamic_settings):
    user = Mock(is_authenticated=True, joined_on=timezone.now())
    cutoff_date = get_cutoff_date(dynamic_settings, user)
    assert cutoff_date == user.joined_on


@override_dynamic_settings(readtracker_cutoff=5)
def test_cutoff_date_for_old_user_is_calculated_from_setting(dynamic_settings):
    user = Mock(is_authenticated=True, joined_on=timezone.now() - timedelta(days=6))
    cutoff_date = get_cutoff_date(dynamic_settings, user)
    valid_cutoff_date = timezone.now() - timedelta(
        days=dynamic_settings.readtracker_cutoff
    )
    assert cutoff_date < valid_cutoff_date
    assert cutoff_date > user.joined_on


def test_cutoff_date_for_anonymous_user_is_calculated_from_setting(dynamic_settings):
    user = Mock(is_authenticated=False)
    cutoff_date = get_cutoff_date(dynamic_settings, user)
    valid_cutoff_date = timezone.now() - timedelta(
        days=dynamic_settings.readtracker_cutoff
    )
    assert cutoff_date < valid_cutoff_date



from django.core import mail as djmail
from django.template.loader import render_to_string
from django.utils.translation import get_language

from ..conf import settings
from .utils import get_host_from_address


def build_mail(recipient, subject, template, sender=None, context=None):
    context = context.copy() if context else {}
    if not context.get("settings"):
        raise ValueError("settings key is missing from context")

    forum_address = context["settings"].forum_address

    context.update(
        {
            "LANGUAGE_CODE": get_language()[:2],
            "LOGIN_URL": settings.LOGIN_URL,
            "forum_host": get_host_from_address(forum_address),
            "user": recipient,
            "sender": sender,
            "subject": subject,
        }
    )

    message_plain = render_to_string("%s.txt" % template, context)
    message_html = render_to_string("%s.html" % template, context)

    message = djmail.EmailMultiAlternatives(
        subject, message_plain, to=[recipient.email]
    )
    message.attach_alternative(message_html, "text/html")

    return message


def mail_user(recipient, subject, template, sender=None, context=None):
    message = build_mail(recipient, subject, template, sender, context)
    message.send()


def mail_users(recipients, subject, template, sender=None, context=None):
    messages = []

    for recipient in recipients:
        messages.append(build_mail(recipient, subject, template, sender, context))

    if messages:
        send_messages(messages)


def send_messages(messages):
    connection = djmail.get_connection()
    connection.send_messages(messages)



from datetime import timedelta
from math import ceil

from django.conf import settings
from django.db import models
from django.utils import timezone

from ...plugins.models import PluginDataModel


class Poll(PluginDataModel):
    category = models.ForeignKey("misago_categories.Category", on_delete=models.CASCADE)
    thread = models.OneToOneField("misago_threads.Thread", on_delete=models.CASCADE)
    poster = models.ForeignKey(
        settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.SET_NULL
    )
    poster_name = models.CharField(max_length=255)
    poster_slug = models.CharField(max_length=255)

    posted_on = models.DateTimeField(default=timezone.now)
    length = models.PositiveIntegerField(default=0)

    question = models.CharField(max_length=255)
    choices = models.JSONField()
    allowed_choices = models.PositiveIntegerField(default=1)
    allow_revotes = models.BooleanField(default=False)

    votes = models.PositiveIntegerField(default=0)
    is_public = models.BooleanField(default=False)

    def move(self, thread):
        if self.thread_id != thread.id:
            self.thread = thread
            self.category_id = thread.category_id
            self.save()

            self.pollvote_set.update(thread=self.thread, category_id=self.category_id)

    @property
    def ends_on(self):
        if self.length:
            return self.posted_on + timedelta(days=self.length)

    @property
    def is_over(self):
        if self.length:
            return timezone.now() > self.ends_on
        return False

    @property
    def thread_type(self):
        return self.category.thread_type

    def get_api_url(self):
        return self.thread_type.get_poll_api_url(self)

    def get_votes_api_url(self):
        return self.thread_type.get_poll_votes_api_url(self)

    def make_choices_votes_aware(self, user):
        if user.is_anonymous:
            for choice in self.choices:
                choice["selected"] = False
            return

        queryset = self.pollvote_set.filter(voter=user).values("choice_hash")
        user_votes = [v["choice_hash"] for v in queryset]

        for choice in self.choices:
            choice["selected"] = choice["hash"] in user_votes

    @property
    def has_selected_choices(self):
        for choice in self.choices:
            if choice.get("selected"):
                return True
        return False

    @property
    def view_choices(self):
        view_choices = []
        for choice in self.choices:
            if choice["votes"] and self.votes:
                proc = int(ceil(choice["votes"] * 100 / self.votes))
            else:
                proc = 0

            view_choices.append(
                {
                    "label": choice["label"],
                    "votes": choice["votes"],
                    "selected": choice["selected"],
                    "proc": proc,
                }
            )
        return view_choices



import pytest
from django.template import Context, Template

from ..outlets import PluginOutletHook, template_outlets


@pytest.fixture
def patch_outlets():
    try:
        org_outlets = template_outlets.copy()
        for key in template_outlets:
            template_outlets[key] = PluginOutletHook()
        yield template_outlets
    finally:
        for key, hook in org_outlets.items():
            template_outlets[key] = hook


@pytest.fixture
def render_outlet_template():
    def render_outlet_template_function(context: dict | None = None):
        template = Template(
            """
            {% load misago_plugins %}
            <div>{% pluginoutlet TEST %}</div>
            """
        )

        return template.render(Context(context or {})).strip()

    return render_outlet_template_function



from django.urls import reverse

from ....notifications.models import Notification


def test_notifications_api_returns_403_error_if_client_is_no_authenticated(db, client):
    response = client.get(reverse("misago:apiv2:notifications"))
    assert response.status_code == 403


def test_notifications_api_returns_empty_list_if_user_has_no_notifications(user_client):
    response = user_client.get(reverse("misago:apiv2:notifications"))
    assert response.status_code == 200
    assert response.json() == {
        "results": [],
        "hasNext": False,
        "hasPrevious": False,
        "firstCursor": None,
        "lastCursor": None,
        "unreadNotifications": None,
    }


def test_notifications_api_returns_list_with_all_user_notifications(user, user_client):
    read_notification = Notification.objects.create(
        user=user, verb="TEST", is_read=True
    )
    notification = Notification.objects.create(user=user, verb="TEST", is_read=False)

    response = user_client.get(reverse("misago:apiv2:notifications"))
    assert response.status_code == 200

    response_json = response.json()
    assert [result["id"] for result in response_json["results"]] == [
        notification.id,
        read_notification.id,
    ]
    assert not response_json["hasNext"]
    assert not response_json["hasPrevious"]


def test_notifications_api_returns_list_with_read_user_notifications(user, user_client):
    read_notification = Notification.objects.create(
        user=user, verb="TEST", is_read=True
    )
    Notification.objects.create(user=user, verb="TEST", is_read=False)

    response = user_client.get(reverse("misago:apiv2:notifications") + "?filter=read")
    assert response.status_code == 200

    response_json = response.json()
    assert [result["id"] for result in response_json["results"]] == [
        read_notification.id
    ]
    assert not response_json["hasNext"]
    assert not response_json["hasPrevious"]


def test_notifications_api_returns_list_with_unread_user_notifications(
    user, user_client
):
    Notification.objects.create(user=user, verb="TEST", is_read=True)
    notification = Notification.objects.create(user=user, verb="TEST", is_read=False)

    response = user_client.get(reverse("misago:apiv2:notifications") + "?filter=unread")
    assert response.status_code == 200

    response_json = response.json()
    assert [result["id"] for result in response_json["results"]] == [notification.id]
    assert not response_json["hasNext"]
    assert not response_json["hasPrevious"]


def test_notifications_api_returns_list_with_notification_by_actor(
    user, other_user, user_client
):
    notification = Notification.objects.create(
        user=user,
        actor=other_user,
        actor_name=other_user.username,
        verb="TEST",
        is_read=False,
    )

    response = user_client.get(reverse("misago:apiv2:notifications"))
    assert response.status_code == 200

    response_json = response.json()
    assert [result["id"] for result in response_json["results"]] == [notification.id]
    assert not response_json["hasNext"]
    assert not response_json["hasPrevious"]


def test_notifications_api_excludes_other_users_notifications(
    user, other_user, user_client
):
    Notification.objects.create(user=other_user, verb="TEST", is_read=True)
    notification = Notification.objects.create(user=user, verb="TEST", is_read=False)

    response = user_client.get(reverse("misago:apiv2:notifications"))
    assert response.status_code == 200

    response_json = response.json()
    assert [result["id"] for result in response_json["results"]] == [notification.id]
    assert not response_json["hasNext"]
    assert not response_json["hasPrevious"]


def test_notifications_api_supports_limiting_results_count(user, user_client):
    Notification.objects.create(user=user, verb="TEST", is_read=False)
    Notification.objects.create(user=user, verb="TEST", is_read=False)
    Notification.objects.create(user=user, verb="TEST", is_read=False)

    response = user_client.get(reverse("misago:apiv2:notifications") + "?limit=2")
    assert response.status_code == 200

    response_json = response.json()
    assert len(response_json["results"]) == 2
    assert response_json["hasNext"]
    assert not response_json["hasPrevious"]


def test_notifications_api_returns_400_error_if_too_many_results_are_requested(
    user, user_client
):
    response = user_client.get(reverse("misago:apiv2:notifications") + "?limit=2000")
    assert response.status_code == 400


def test_notifications_api_clears_unread_notifications_if_user_has_no_notifications(
    user, user_client
):
    user.unread_notifications = 10
    user.save()

    response = user_client.get(reverse("misago:apiv2:notifications"))
    assert response.status_code == 200

    response_json = response.json()
    assert not response_json["results"]
    assert not response_json["hasNext"]
    assert not response_json["hasPrevious"]
    assert response_json["unreadNotifications"] is None

    user.refresh_from_db()
    assert user.unread_notifications == 0


def test_notifications_api_clears_unread_notifications_if_unread_list_is_empty(
    user, user_client
):
    user.unread_notifications = 10
    user.save()

    response = user_client.get(reverse("misago:apiv2:notifications") + "?filter=unread")
    assert response.status_code == 200

    response_json = response.json()
    assert not response_json["results"]
    assert not response_json["hasNext"]
    assert not response_json["hasPrevious"]
    assert response_json["unreadNotifications"] is None

    user.refresh_from_db()
    assert user.unread_notifications == 0


def test_notifications_api_recounts_unread_notifications_if_only_page_has_unread_items(
    user, user_client
):
    user.unread_notifications = 0
    user.save()

    Notification.objects.create(user=user, verb="TEST", is_read=False)

    response = user_client.get(reverse("misago:apiv2:notifications"))
    assert response.status_code == 200

    response_json = response.json()
    assert response_json["results"]
    assert response_json["unreadNotifications"] == "1"

    user.refresh_from_db()
    assert user.unread_notifications == 1


def test_notifications_api_recounts_unread_notifications_if_unread_list_has_items(
    user, user_client
):
    user.unread_notifications = 0
    user.save()

    notification = Notification.objects.create(user=user, verb="TEST", is_read=False)

    response = user_client.get(
        reverse("misago:apiv2:notifications")
        + f"?filter=unread&before={notification.id - 1}"
    )
    assert response.status_code == 200

    response_json = response.json()
    assert response_json["results"]
    assert response_json["unreadNotifications"] == "1"

    user.refresh_from_db()
    assert user.unread_notifications == 1


def test_notifications_api_recounts_unread_notifications_if_user_has_new_notifications(
    user, user_client
):
    user.unread_notifications = 0
    user.save()

    notification = Notification.objects.create(user=user, verb="TEST", is_read=False)

    response = user_client.get(
        reverse("misago:apiv2:notifications") + f"?before={notification.id - 1}"
    )
    assert response.status_code == 200

    response_json = response.json()
    assert response_json["results"]
    assert response_json["unreadNotifications"] == "1"

    user.refresh_from_db()
    assert user.unread_notifications == 1



from django.urls import reverse

from ...conf.test import override_dynamic_settings


def test_view_begins_social_auth_for_provider(client, provider):
    response = client.get(
        reverse("misago:social-begin", kwargs={"backend": provider.pk})
    )
    assert response.status_code == 302


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_view_returns_403_when_oauth2_is_enabled(client, provider):
    response = client.get(
        reverse("misago:social-begin", kwargs={"backend": provider.pk})
    )
    assert response.status_code == 403


def test_view_returns_404_for_disabled_provider(client, disabled_provider):
    response = client.get(
        reverse("misago:social-begin", kwargs={"backend": disabled_provider.pk})
    )
    assert response.status_code == 404


def test_view_returns_404_for_undefined_provider(db, client):
    response = client.get(
        reverse("misago:social-begin", kwargs={"backend": "undefined"})
    )
    assert response.status_code == 404



from unittest.mock import Mock

from django.core.management import call_command


def test_management_command_invalidates_all_caches(mocker):
    invalidate_all_caches = mocker.patch("misago.cache.versions.invalidate_all_caches")
    call_command("invalidateversionedcaches", stdout=Mock())
    invalidate_all_caches.assert_called_once()



from django.db import migrations


class Migration(migrations.Migration):
    """Migration superseded by 0016"""

    dependencies = [
        ("misago_users", "0002_users_settings"),
        ("misago_core", "0001_initial"),
    ]

    operations = []



from ..models import WatchedThread
from ..threads import ThreadNotifications, watch_replied_thread


def test_replied_thread_is_watched_with_email_notifications(user, thread):
    user.watch_replied_threads = ThreadNotifications.SITE_AND_EMAIL
    user.save()

    watch_replied_thread(user, thread)

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert watched_thread.send_emails


def test_replied_thread_is_watched_without_email_notifications(user, thread):
    user.watch_replied_threads = ThreadNotifications.SITE_ONLY
    user.save()

    watch_replied_thread(user, thread)

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert not watched_thread.send_emails


def test_replied_thread_is_not_watched_if_option_is_disabled(user, thread):
    user.watch_replied_threads = ThreadNotifications.NONE
    user.save()

    watch_replied_thread(user, thread)

    assert not WatchedThread.objects.exists()


def test_replied_thread_watching_entry_notification_emails_are_not_disabled(
    user, thread, watched_thread_factory
):
    user.watch_replied_threads = ThreadNotifications.SITE_ONLY
    user.save()

    watched_thread = watched_thread_factory(user, thread, send_emails=True)

    watch_replied_thread(user, thread)

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert watched_thread.send_emails


def test_replied_thread_watching_entry_notification_emails_are_not_enabled(
    user, thread, watched_thread_factory
):
    user.watch_replied_threads = ThreadNotifications.SITE_AND_EMAIL
    user.save()

    watched_thread = watched_thread_factory(user, thread, send_emails=False)

    watch_replied_thread(user, thread)

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert not watched_thread.send_emails


def test_replied_thread_watching_entry_is_not_removed_if_option_is_disabled(
    user, thread, watched_thread_factory
):
    user.watch_replied_threads = ThreadNotifications.NONE
    user.save()

    watched_thread = watched_thread_factory(user, thread, send_emails=False)

    watch_replied_thread(user, thread)

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert not watched_thread.send_emails



# Generated by Django 4.2.7 on 2023-12-05 19:22

import django.contrib.postgres.indexes
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("misago_users", "0025_alter_user_rank"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="plugin_data",
            field=models.JSONField(default=dict),
        ),
        migrations.AddIndex(
            model_name="user",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["plugin_data"], name="misago_user_plugin__c45a32_gin"
            ),
        ),
    ]



from time import time

from django.core.exceptions import PermissionDenied
from django.urls import reverse
from django.utils.translation import pgettext
from rest_framework.decorators import api_view
from rest_framework.response import Response

from ..core.shortcuts import get_int_or_404
from .searchproviders import searchproviders


@api_view()
def search(request, search_provider=None):
    allowed_providers = searchproviders.get_allowed_providers(request)
    if not request.user_acl["can_search"] or not allowed_providers:
        raise PermissionDenied(
            pgettext("search api", "You don't have permission to search site.")
        )

    search_query = get_search_query(request)
    response = []
    for provider in allowed_providers:
        provider_data = {
            "id": provider.url,
            "name": str(provider.name),
            "icon": provider.icon,
            "url": reverse("misago:search", kwargs={"search_provider": provider.url}),
            "api": reverse(
                "misago:api:search", kwargs={"search_provider": provider.url}
            ),
            "results": None,
            "time": None,
        }

        if not search_provider or search_provider == provider.url:
            start_time = time()

            if search_provider == provider.url:
                page = get_int_or_404(request.query_params.get("page", 1))
            else:
                page = 1

            provider_data["results"] = provider.search(search_query, page)
            provider_data["time"] = float("%.2f" % (time() - start_time))

        response.append(provider_data)
    return Response(response)


def get_search_query(request):
    return request.query_params.get("q", "").strip()



from django.conf import settings
from django.db import IntegrityError, models
from django.utils import timezone


class Online(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        primary_key=True,
        related_name="online_tracker",
        on_delete=models.CASCADE,
    )
    last_click = models.DateTimeField(default=timezone.now)

    def save(self, *args, **kwargs):
        try:
            super().save(*args, **kwargs)
        except IntegrityError:
            pass  # first come is first serve in online tracker
